import logging
import os

import iridauploader.progress as progress

from iridauploader.parsers import exceptions
from iridauploader.parsers import common
from iridauploader.parsers import BaseParser
from iridauploader.parsers.directory import sample_parser, validation


class Parser(BaseParser):

    SAMPLE_SHEET_FILE_NAME = 'SampleList.csv'

    def __init__(self, parser_type_name='directory'):
        """
        Initialize the Parser
        :param parser_type_name: string to be included in metadata of sequencing run for type identification in IRIDA
        """
        super().__init__(parser_type_name=parser_type_name, required_file_list=[Parser.SAMPLE_SHEET_FILE_NAME])

    def find_runs(self, directory):
        """
        find a list of run directories in the directory given

        :param directory:
        :return: list of DirectoryStatus objects
        """
        logging.info("Looking for runs in {}".format(directory))

        runs = []
        directory_list = common.find_directory_list(directory)
        for d in directory_list:
            runs.append(progress.get_directory_status(d, self.get_required_file_list()))

        return runs

    def find_single_run(self, directory):
        """
        Find a run in the base directory given

        :param directory:
        :return: DirectoryStatus object
        """
        logging.info("looking for run in {}".format(directory))

        return progress.get_directory_status(directory, self.get_required_file_list())

    @staticmethod
    def get_sample_sheet(directory):
        """
        gets the sample sheet file path from a given run directory

        :param directory:
        :return:
        """
        logging.info("Looking for sample sheet in {}".format(directory))

        # Checks if we can access to the given directory, return empty and log a warning if we cannot.
        if common.cannot_read_directory(directory):
            logging.error(("The directory is not accessible, can not parse samples from this directory {}"
                           "".format(directory), directory))
            raise exceptions.DirectoryError("The directory is not accessible, "
                                            "can not parse samples from this directory {}".format(directory),
                                            directory)

        sample_sheet_file_name = Parser.SAMPLE_SHEET_FILE_NAME
        file_list = common.get_file_list(directory)
        if sample_sheet_file_name not in file_list:
            logging.error("No sample sheet file in the Directory Upload format found")
            raise exceptions.DirectoryError("The directory {} has no sample sheet file in the Directory Upload format "
                                            "with the name {}"
                                            "".format(directory, sample_sheet_file_name), directory)
        else:
            logging.debug("Sample sheet found")
            return os.path.join(directory, sample_sheet_file_name)

    def get_sequencing_run(self, sample_sheet, run_data_directory_file_list=None):
        """
        Does local validation on the integrity of the run directory / sample sheet

        Throws a ValidationError with a validation result attached if it cannot make a sequencing run

        :param sample_sheet:
        :param run_data_directory_file_list: Optional: List of files in the data directory to verify against the
        SampleList.csv file. This is used when deploying the parsers on a cloud environment.
        :return: SequencingRun
        """

        # Try to get the sample sheet, validate that the sample sheet is valid
        validation_result = validation.validate_sample_sheet(sample_sheet)
        if not validation_result.is_valid():
            logging.error("Errors occurred while getting sample sheet")
            raise exceptions.ValidationError("Errors occurred while getting sample sheet", validation_result)

        # When running with a premade file list, verify files on sample_sheet are in file list
        try:
            if run_data_directory_file_list is not None:
                sample_parser.verify_sample_sheet_file_names_in_file_list(sample_sheet, run_data_directory_file_list)
        except (exceptions.SequenceFileError, exceptions.SampleSheetError) as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while building sequence run from sample sheet")
            raise exceptions.ValidationError("Errors occurred while building sequence run from sample sheet",
                                             validation_result)
        except Exception as error:
            validation_result.add_error(error)
            logging.error("System error while building sequencing run")
            raise exceptions.ValidationError("System error while building sequencing run", validation_result)

        # Build a list of sample objects from sample sheet
        try:
            if run_data_directory_file_list is not None:
                sample_list = sample_parser.build_sample_list_from_sample_sheet_no_verify(sample_sheet)
            else:
                sample_list = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(sample_sheet)
        except (exceptions.DirectoryError, exceptions.SampleSheetError) as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while parsing files")
            raise exceptions.ValidationError("Errors occurred while parsing files", validation_result)
        except Exception as error:
            validation_result.add_error(error)
            logging.error("System error while parsing files")
            raise exceptions.ValidationError("System error while parsing files", validation_result)

        # verify samples in sample_list are all of one type, either single or paired end
        if not sample_parser.only_single_or_paired_in_sample_list(sample_list):
            e = exceptions.SampleSheetError(
                ("Your sample sheet is malformed. "
                 "SampleSheet cannot have both paired end and single end runs. "
                 "Make sure all samples are either paired or single."),
                sample_sheet
            )
            validation_result.add_error(e)
            logging.error("Error occurred while building file list: Sample sheet has both paired and single end reads")
            raise exceptions.ValidationError("Errors occurred while building file list.", validation_result)

        # Try to build sequencing run from sample sheet & meta data, raise validation error if errors occur
        try:
            run_metadata = sample_parser.parse_metadata(sample_list)
            sequencing_run = common.build_sequencing_run_from_samples(sample_list,
                                                                      run_metadata,
                                                                      self.get_parser_type_name())
        except exceptions.SequenceFileError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while building sequence run from sample sheet")
            raise exceptions.ValidationError("Errors occurred while building sequence run from sample sheet",
                                             validation_result)

        return sequencing_run



import logging
# PyQt needs to be imported like this because for whatever reason they decided not to include a __all__ = [...]
import PyQt5.QtWidgets as QtWidgets
import PyQt5.QtCore as QtCore
import PyQt5.QtGui as QtGui

from iridauploader.api import UPLOAD_MODES
from iridauploader.config import config
from iridauploader.model import DirectoryStatus

from iridauploader.gui.config import ConfigDialog
from iridauploader.gui import tools, widgets, colours, threads

import os


class MainDialog(QtWidgets.QDialog):
    """
    The programs main interactive window.

    Includes directory selection, automatic parsing, sample/project preview and error reporting.

    The user will be warned when a run is invalid, has already been uploaded, or errors happen during upload.
    """
    def __init__(self):
        super().__init__()
        logging.debug("GUI: Setting up MainDialog")

        self.config_dlg = ConfigDialog(parent=self)

        # set window title and icon
        self.setWindowTitle("IRIDA Uploader")
        icon_path = os.path.join(os.path.dirname(__file__), 'images/icon.ico')
        self.setWindowIcon(QtGui.QIcon(icon_path))

        # Initialize threads from the tools file
        self._status_thread = threads.StatusThread()
        self._parse_thread = threads.ParseThread()
        self._upload_thread = threads.UploadThread()

        # internal variables
        self._run_dir = ""
        self._config_file = ""
        self._uploading = False
        self._continue_partial = False

        # Setup gui objects
        self._init_objects()

        # Set Layout and Geometry
        self.setLayout(self._init_layout())
        # resize window
        self.setGeometry(0, 0, 800, 600)
        # Center Window
        qt_rectangle = self.frameGeometry()
        center_point = QtWidgets.QDesktopWidget().availableGeometry().center()
        qt_rectangle.moveCenter(center_point)
        self.move(qt_rectangle.topLeft())

        # Signals and Slots
        # buttons
        self._dir_button.clicked.connect(self._btn_open_dir)
        self._config_button.clicked.connect(self._btn_show_config)
        self._refresh_button.clicked.connect(self._btn_refresh)
        self._upload_button.clicked.connect(self._btn_upload)
        self._console_button.clicked.connect(self._btn_log)
        self._info_btn.clicked.connect(self._btn_continue)
        self._info_partial_btn.clicked.connect(self._btn_continue_partial)
        # connect threads finishing to finish functions
        self._status_thread.finished.connect(self._thread_finished_status)
        self._parse_thread.finished.connect(self._thread_finished_parse)
        self._upload_thread.finished.connect(self._thread_finished_upload)

        # init the config file
        config.setup()
        # block uploading at start
        self._upload_button.set_block()

    def _init_objects(self):
        """
        Setup all the objects that appear in the program
        :return: None
        """
        # directory
        self._dir_button = QtWidgets.QPushButton(self)
        self._dir_button.setText("Open Run Directory")
        self._dir_button.setFixedWidth(200)
        self._dir_line = QtWidgets.QLineEdit(self)
        self._dir_line.setReadOnly(True)
        self._dir_line.setEnabled(False)
        self._dir_line.setStyleSheet("color: black")
        # config
        self._config_button = QtWidgets.QPushButton(self)
        self._config_button.setText("Configure Settings")
        self._config_button.setFixedWidth(200)
        # refresh
        self._refresh_button = QtWidgets.QPushButton(self)
        self._refresh_button.setText("Refresh")
        # upload mode
        self._upload_mode_label = QtWidgets.QLabel("Upload Mode: ")
        self._upload_mode_combobox = QtWidgets.QComboBox(self)
        self._upload_mode_combobox.addItems(UPLOAD_MODES)
        self._upload_mode_combobox.setCurrentIndex(0)
        # Info lines, these start out as hidden
        self._info_line = QtWidgets.QLineEdit(self)
        self._info_line.setReadOnly(True)
        self._info_line.setStyleSheet("background-color: {}".format(colours.YELLOW_LIGHT))
        self._info_line.hide()
        self._prev_errors = QtWidgets.QPlainTextEdit(self)
        self._prev_errors.setReadOnly(True)
        self._prev_errors.hide()
        self._info_btn = QtWidgets.QPushButton(self)
        self._info_btn.setText("Continue")
        self._info_btn.setStyleSheet("background-color: {}".format(colours.RED_LIGHT))
        self._info_btn.hide()
        self._info_partial_btn = QtWidgets.QPushButton(self)
        self._info_partial_btn.setText("Continue Partially Uploaded Run")
        self._info_partial_btn.setStyleSheet("background-color: {}".format(colours.BLUE_LIGHT))
        self._info_partial_btn.hide()
        self._curr_errors = QtWidgets.QPlainTextEdit(self)
        self._curr_errors.setReadOnly(True)
        self._curr_errors.setStyleSheet("background-color: {}".format(colours.RED_LIGHT))
        self._curr_errors.hide()
        # Upload button
        self._upload_button = widgets.UploadButton(self)
        # Table
        self._table = widgets.SampleTable(self)
        # Upload error text
        self._upload_errors = QtWidgets.QPlainTextEdit(self)
        self._upload_errors.setReadOnly(True)
        self._upload_errors.setStyleSheet("background-color: {}".format(colours.RED_LIGHT))
        self._upload_errors.hide()
        # Logging console
        self._console = widgets.LogTextBox(self)
        self._console_button = QtWidgets.QPushButton(self)
        self._console_button.setFixedWidth(100)
        self._console_button.setText("Show Log")

    def _init_layout(self):
        """
        Setup layout
        :return: QtWidgets.QVBoxLayout
        """
        # main layout
        layout = QtWidgets.QVBoxLayout()

        # Directory selection
        dir_layout = QtWidgets.QHBoxLayout()
        dir_layout.addWidget(self._dir_button)
        dir_layout.addWidget(self._dir_line)
        layout.addLayout(dir_layout)

        # Config selection & upload mode & refresh
        config_layout = QtWidgets.QHBoxLayout()
        config_layout.addWidget(self._config_button)
        upload_mode_layout = QtWidgets.QHBoxLayout()
        upload_mode_layout.addWidget(self._upload_mode_label)
        upload_mode_layout.addWidget(self._upload_mode_combobox)
        config_layout.addLayout(upload_mode_layout)
        config_layout.addWidget(self._refresh_button)
        layout.addLayout(config_layout)

        # info
        layout.addWidget(self._info_line)
        layout.addWidget(self._prev_errors)
        info_btn_layout = QtWidgets.QHBoxLayout()
        info_btn_layout.addWidget(self._info_btn)
        info_btn_layout.addWidget(self._info_partial_btn)
        layout.addLayout(info_btn_layout)
        layout.addWidget(self._curr_errors)

        # table
        layout.addWidget(self._table)

        # Upload error
        layout.addWidget(self._upload_errors)

        # Upload button
        upload_layout = QtWidgets.QHBoxLayout()
        upload_layout.addWidget(self._upload_button)
        upload_layout.addWidget(self._console_button)
        layout.addLayout(upload_layout)

        # logging text box
        layout.addWidget(self._console)

        return layout

    def closeEvent(self, event):
        """
        Override closeEvent to give warning before closing the application.

        If uploads are currently running, user will prompted before exiting the program
        :param event:
        :return:
        """
        if self._uploading:
            reply = QtWidgets.QMessageBox.question(self, 'Exit Uploader?',
                                                   "Data is still being uploaded! "
                                                   "Are you sure you want to exit the program?",
                                                   QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
            if reply == QtWidgets.QMessageBox.Yes:
                event.accept()
            else:
                event.ignore()
        else:
            event.accept()

    #################
    #    Buttons    #
    #################

    def _btn_open_dir(self):
        """
        Opens a file(directory) dialog and sets the _run_dir variable to chosen directory
        Updates ui to reflect changes
        :return:
        """
        logging.debug("GUI: _btn_open_dir clicked")
        self._run_dir = str(QtWidgets.QFileDialog.getExistingDirectory(self, "Select Directory"))
        self._dir_line.setText(self._run_dir)
        logging.debug("GUI: result: " + self._run_dir)

        # Kick of status check and parsing after opening a new directory
        self._start_status()

    def _btn_show_config(self):
        """
        Open the config dialog window
        After the dialog closes, contacts irida again (which updates the gui element)
        :return:
        """
        self.config_dlg.center_window()
        self.config_dlg.exec()

        # restart the status thread
        self._start_status()

    def _btn_refresh(self):
        """
        Clicking the refresh button re-tries the connection to IRIDA, and then re-tries a parse
        :return:
        """
        # restart the status thread
        self._start_status()

    def _btn_upload(self):
        """
        Blocks usage of the ui elements, and starts the upload thread with the set variables
        :return:
        """
        logging.debug("GUI: _btn_upload clicked")

        try:
            tools.is_connected_to_irida()
            self._start_upload()
        except Exception as e:
            # If no connection to irida, display a warning to the user
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText("Cannot Connect to IRIDA: " + str(e))
            msg.setInformativeText('Please check your configuration')
            msg.setWindowTitle("Connection Error")
            # You need 2 buttons on the widget for the exit button to not be the same as the Open Config button
            # So we add an "OK" button, and then hide it
            p_config_button = msg.addButton("Open Config", QtWidgets.QMessageBox.YesRole)
            p_accept_button = msg.addButton("OK", QtWidgets.QMessageBox.NoRole)
            p_accept_button.hide()
            # Show the window, and don't let user interact with main dialog
            msg.exec()
            # allow user to go directoy to the config page
            if msg.clickedButton() == p_config_button:
                self._btn_show_config()

    def _btn_log(self):
        """
        If the log is hidden, show, and vice versa
        :return:
        """
        if self._console.isVisible():
            self._console.hide()
            self._console_button.setText("Show Log")
        else:
            self._console.show()
            self._console_button.setText("Hide Log")

    def _btn_continue(self, continue_partial=False):
        """
        Reset the error gui elements, and continue onto the parse phase
        By default does not use continue_partial flag
        :return:
        """
        self._continue_partial = continue_partial
        self._reset_previous_error()
        self._reset_info_line()
        self._start_parse()

    def _btn_continue_partial(self):
        """
        Continues the upload with the continue_partial flag enabled
        :return:
        """
        self._btn_continue(continue_partial=True)

    #######################
    #   Thread Starters   #
    #######################

    def _start_status(self):
        """
        Blocks usage of the ui elements
        Clear data and info out of gui
        Starts the status thread
        This acts as the logical start point for status/parsing run directories
        :return:
        """
        logging.debug("GUI: _btn_upload clicked")

        # lock the gui so users don't click things while the parsing is still happening.
        self._lock_gui()
        # Clear everything
        self._table.clear_table()
        self._reset_info_line()
        self._reset_previous_error()
        self._reset_current_error()
        self._reset_upload_error()

        # start status thread
        self._status_thread.set_vars(self._run_dir)
        self._status_thread.start()

    def _start_parse(self):
        """
        Starts the parse thread
        :return:
        """
        logging.debug("GUI: _start_parse clicked")
        # lock gui
        self._lock_gui()
        # start parsing
        self._parse_thread.set_vars(self._run_dir, self._continue_partial)
        self._parse_thread.start()

    def _start_upload(self):
        # Lock Gui
        self._lock_gui()
        self._upload_button.set_uploading()
        self._uploading = True
        # start upload
        upload_mode = self._upload_mode_combobox.currentText()
        self._upload_thread.set_vars(
            run_dir=self._run_dir, upload_mode=upload_mode, partial_continue=self._continue_partial)
        self._upload_thread.start()

    ##########################
    #    Threads Complete    #
    ##########################

    def _thread_finished_status(self):
        """
        Main logic for post status

        unlock gui (blocked when thread started)

        On invalid runs:
            show user error
            don't let user continue past error to parsing
            block uploading (no continue allowed)
        On New runs:
            start parsing run directory
        On Complete / Partial / Delayed / Error runs:
            Show user the state (complete, partial, delayed, error) and the reason for the state (error msg)
            Allow users to click continue to continue on to parsing the run
            Block uploading (until continue is clicked)

        :return: None
        """
        logging.debug("GUI: _thread_finished_status called")

        self._continue_partial = False

        # since the thread finished, we need to unlock the gui
        self._unlock_gui()

        status = self._status_thread.get_result()

        if status.status_equals(DirectoryStatus.INVALID):
            # Then we need to block upload, since it's invalid
            self._upload_button.set_block()
            # Give info in info line
            self._show_and_fill_info_line("Run is not valid: " + str(status.message))
            # an invalid run cannot be continued
            self._hide_info_button()

        elif status.status_equals(DirectoryStatus.NEW):
            # new runs start the parse immediately
            self._start_parse()

        elif status.status_equals(DirectoryStatus.COMPLETE):
            # We need to block upload until the user clicks continue
            self._upload_button.set_block()
            # give user info
            self._show_and_fill_info_line("This run directory has already been uploaded. "
                                          "Click 'Continue' if you want to proceed anyway.")

        elif status.status_equals(DirectoryStatus.DELAYED):
            # We need to block upload until the user clicks continue
            self._upload_button.set_block()
            # give user info
            self._show_and_fill_info_line("This run directory has been given a delay. "
                                          "Click 'Continue' if you want to proceed anyway.")

        elif status.status_equals(DirectoryStatus.PARTIAL):
            # We need to block upload until the user clicks continue
            self._upload_button.set_block()
            # give user info
            user_message = "This run directory is partially uploaded. Choose how you would like to Continue."
            if status.message is not None:
                user_message = user_message + " This directory had the error(s) below. "
                self._show_previous_error(status.message)
            self._show_and_fill_info_partial_upload_options(user_message)

        elif status.status_equals(DirectoryStatus.ERROR):
            # We need to block upload until the user clicks continue
            self._upload_button.set_block()
            # give user info
            self._show_and_fill_info_line("This run directory previously had the error(s) below. "
                                          "Click 'Continue' if you want to proceed anyway.")
            self._show_previous_error(status.message)

    def _thread_finished_parse(self):
        """
        Checks if run was parsed correctly or not,
        fills table is valid
        shows users error and blocks upload if run did not parse
        Unblocks the UI elements
        :return:
        """
        logging.debug("GUI: _thread_finished_parse called")

        # get data from the thread object
        sequencing_run = self._parse_thread.get_run()

        if sequencing_run:
            # run parsed correctly
            self._table.fill_table(sequencing_run, self._continue_partial)
            self._unlock_gui()
        else:
            run_errors = self._parse_thread.get_error()
            self._show_current_error(run_errors)
            self._unlock_gui()
            # block uploading runs with parsing errors
            self._upload_button.set_block()

    def _thread_finished_upload(self):
        """
        Unblocks the UI elements,
        :return:
        """
        self._uploading = False
        self._continue_partial = False

        logging.debug("GUI: _thread_finished_upload called")
        if not self._upload_thread.is_success():
            error = self._upload_thread.get_exit_error()
            self._show_upload_error(str(error))
            # Unlock GUI
            self._unlock_gui()
            self._upload_button.set_block()
        else:
            # Unlock GUI
            self._unlock_gui()
            self._upload_button.set_finished()

    ############################
    #   Show/Hide/Fill/Clear   #
    ############################

    def _show_and_fill_info_line(self, message):
        """
        shows the info line
        fills the info line with message
        shows the info button (continue button)
        :param message: string to display to the user
        :return:
        """
        self._info_line.setText(message)
        self._info_line.show()
        self._info_btn.show()

    def _show_and_fill_info_partial_upload_options(self, message):
        """
        When it is a partial run, show info line with multiple options

        :param message: string to display to the user
        :return:
        """
        self._info_line.setText(message)
        self._info_line.show()
        self._info_btn.setText("Restart Run From Beginning")
        self._info_btn.show()
        self._info_partial_btn.show()

    def _reset_info_line(self):
        """
        Hides the info line
        blanks out the info line
        :return:
        """
        self._info_line.setText("")
        self._info_line.hide()
        self._info_btn.setText("Continue")
        self._info_btn.hide()
        self._info_partial_btn.hide()

    def _hide_info_button(self):
        """
        disables the use of the info (continue) button
        Useful in the case of an invalid run that should never be parsed
        :return:
        """
        self._info_btn.hide()
        self._info_partial_btn.hide()

    def _show_previous_error(self, errors):
        """
        Shows errors from a previous run attempt
        unhides a textbox and fills it with a string
        :param errors: string of errors to display to user
        :return:
        """
        self._prev_errors.show()
        self._prev_errors.appendPlainText(str(errors))

    def _reset_previous_error(self):
        """
        blanks out and hides the previous error box
        :return:
        """
        self._prev_errors.clear()
        self._prev_errors.hide()

    def _show_current_error(self, errors):
        """
        Shows errors from a current parse attempt
        unhides a textbox and fills it with a string
        :param errors: string of errors to display to user
        :return:
        """
        self._curr_errors.show()
        self._curr_errors.appendPlainText(str(errors))

    def _reset_current_error(self):
        """
        blanks out and hides the current error box
        :return:
        """
        self._curr_errors.clear()
        self._curr_errors.hide()

    def _show_upload_error(self, errors):
        """
        Shows errors from a current upload
        unhides a textbox and fills it with a string
        :param errors: string of errors to display to user
        :return:
        """
        self._upload_errors.show()
        self._upload_errors.appendPlainText(str(errors))

    def _reset_upload_error(self):
        """
        blanks out and hides the current upload error box
        :return:
        """
        self._upload_errors.clear()
        self._upload_errors.hide()

    def _lock_gui(self):
        """
        Locks gui elements that users should not be able to interact with when threads are running
        :return:
        """
        self._config_button.setEnabled(False)
        self._dir_button.setEnabled(False)
        self._refresh_button.setEnabled(False)
        self._upload_button.set_block()

    def _unlock_gui(self):
        """
        Unlocks all the gui elements that are blocked for threading reasons
        :return:
        """
        self._config_button.setEnabled(True)
        self._dir_button.setEnabled(True)
        self._refresh_button.setEnabled(True)

        # Check that we have a connection to IRIDA, if there is no connection, block upload to prevent errors
        self._upload_button.set_ready()



from iridauploader.progress.upload_status import get_directory_status, write_directory_status, run_is_ready_with_delay
from iridauploader.progress.upload_signals import signal_worker, send_progress, ProgressData
from iridauploader.progress import exceptions



import unittest
from os import path

from iridauploader.parsers.nextseq.parser import Parser
from iridauploader.parsers.exceptions import DirectoryError, ValidationError, SampleSheetError
import iridauploader.model as model

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestFindRuns(unittest.TestCase):
    """
    Test finding runs in a directory filled with (possible) run directories
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_find_one_of_three(self):
        """
        Given a directory with 3 potential run directories in it, determine which are new and invalid
        :return:
        """
        directory = path.join(path_to_module, "three_dirs")
        dir_1 = path.join(directory, "first")
        dir_2 = path.join(directory, "second")
        dir_3 = path.join(directory, "third")
        correct_dirs = [dir_1, dir_2, dir_3]

        res = Parser().find_runs(directory)

        self.assertIn(res[0].directory, correct_dirs)
        self.assertIn(res[1].directory, correct_dirs)
        self.assertIn(res[2].directory, correct_dirs)

        for c_dir in res:
            if c_dir.directory == dir_1:
                self.assertEqual(c_dir.status, "new")
            elif c_dir.directory == dir_2:
                self.assertEqual(c_dir.status, "invalid")
            elif c_dir.directory == dir_3:
                self.assertEqual(c_dir.status, "invalid")

    def test_find_none(self):
        """
        Given a directory with no run directories in it, return a empty list
        :return:
        """
        directory = path.join(path_to_module, "no_dirs")

        res = Parser().find_runs(directory)

        self.assertEqual(res, [])


class TestFindSingleRun(unittest.TestCase):
    """
    Test checking if a single directory is a run
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_find_in_dir_with_three_dirs(self):
        """
        Given a directory with directories in it (but no actual run at highest level), show as invalid
        :return:
        """
        directory = path.join(path_to_module, "three_dirs")

        res = Parser().find_single_run(directory)

        self.assertEqual(type(res), model.DirectoryStatus)
        self.assertEqual(res.status, "invalid")
        self.assertEqual(res.directory, directory)

    def test_find_none(self):
        """
        Given a directory with nothing in it, show as invalid
        :return:
        """
        directory = path.join(path_to_module, "no_dirs")

        res = Parser().find_single_run(directory)

        self.assertEqual(type(res), model.DirectoryStatus)
        self.assertEqual(res.status, "invalid")
        self.assertEqual(res.directory, directory)

    def test_find_dir(self):
        """
        Given a valid run directory, show as new
        :return:
        """
        directory = path.join(path_to_module, "three_dirs", "first")

        res = Parser().find_single_run(directory)

        self.assertEqual(type(res), model.DirectoryStatus)
        self.assertEqual(res.status, "new")
        self.assertEqual(res.directory, directory)

    def test_find_dir_no_completed_file(self):
        """
        Given a run directory with no run completed file, show as invalid
        :return:
        """
        directory = path.join(path_to_module, "fake_ngs_no_rtacomplete")

        res = Parser().find_single_run(directory)

        self.assertEqual(type(res), model.DirectoryStatus)
        self.assertEqual(res.status, "invalid")
        self.assertEqual(res.directory, directory)


class TestGetSampleSheet(unittest.TestCase):
    """
    Test cases related to getting a sample sheet from a known run directory
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid_directory(self):
        """
        test the file path of a valid sequencing run is gotten correctly
        :return:
        """
        directory = path.join(path_to_module, "three_dirs", "first")
        file_path = path.join(directory, "SampleSheet.csv")

        res = Parser().get_sample_sheet(directory)

        self.assertEqual(res, file_path)

    def test_no_sheet_in_directory(self):
        """
        Given an invalid directory, ensure an error is thrown
        :return:
        """
        directory = path.join(path_to_module, "three_dirs", "third")

        with self.assertRaises(DirectoryError) as context:
            Parser().get_sample_sheet(directory)

        self.assertEqual(context.exception.directory, directory)

    def test_unreadable_directory(self):
        """
        Given an invalid directory (chmod 000), ensure the correct error is thown
        :return:
        """
        directory = path.join(path_to_module, "inaccessible_dir")

        with self.assertRaises(DirectoryError) as context:
            Parser().get_sample_sheet(directory)

        self.assertEqual(context.exception.directory, directory)


class TestGetSequencingRun(unittest.TestCase):
    """
    Test building a sequencing run from a sample sheet
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_invalid_sample_sheets(self):
        """
        Given an invalid sample sheet, ensure the correct errors are included in the validation result
        :return:
        """
        sample_sheet = path.join(path_to_module, "invalid_sample_sheet", "SampleSheet.csv")

        with self.assertRaises(ValidationError) as context:
            Parser().get_sequencing_run(sample_sheet)

        validation_result = context.exception.validation_result
        self.assertEqual(type(validation_result), model.ValidationResult)

        for error in validation_result.error_list:
            self.assertEqual(type(error), SampleSheetError)

    def test_valid_run(self):
        """
        Given a valid sample sheet, ensure a sequencing run is created
        :return:
        """
        sample_sheet = path.join(path_to_module, "fake_nextseq_run", "SampleSheet.csv")

        res = Parser().get_sequencing_run(sample_sheet)

        self.assertEqual(type(res), model.SequencingRun)



from iridauploader.parsers.base_parser import BaseParser
from iridauploader.parsers.parsers import supported_parsers
from iridauploader.parsers.parsers import parser_factory
from iridauploader.parsers import exceptions
from iridauploader.parsers import common



class IridaUploadCanceledException(Exception):
    """
    This error can be thrown mid upload if the api receives the call to cancel uploads

    This should be caught and the information that the upload has been canceled should be relayed
    """
    pass



# todo: an improved code flow may include another modules for handling pubsub, and having the QSignalWorker sub to that
#  that way, the QSignalWorker will only get imported from gui code, and not api code (which would only load the new
#  pubsub module)

import sys
# Check if the PyQt5 is installed
GUI_PKG_INSTALLED = 'PyQt5' in sys.modules

# Don't import gui stuff if we are doing command line only
if GUI_PKG_INSTALLED:
    from PyQt5 import QtCore


class ProgressData:
    """
    A class to wrap upload progress data with standardised getters/setters
    """
    def __init__(self, sample, project, progress):
        self._sample = sample
        self._project = project
        self._progress = progress

    @property
    def sample(self):
        return self._sample

    @property
    def project(self):
        return self._project

    @property
    def progress(self):
        return self._progress


# int to none so send_progress knows it exists
signal_worker = None

# Class does not need to be defined unless gui is loaded
if GUI_PKG_INSTALLED:
    class QSignalWorker(QtCore.QObject):
        # we are sending ProgressData type data
        progress_signal = QtCore.pyqtSignal(ProgressData)

        def send_progress(self, data):
            self.progress_signal.emit(data)

    signal_worker = QSignalWorker()


def send_progress(data):
    # Only send progress if gui is loaded
    if GUI_PKG_INSTALLED:
        signal_worker.send_progress(data)



import re
from os import path
from collections import OrderedDict
from copy import deepcopy
import logging

import iridauploader.model as model
from iridauploader.parsers import exceptions
from iridauploader.parsers import common


def parse_metadata(sample_sheet_file):

    """
    Parse all lines under [Header], [Reads] and [BCLConvert_Settings] in .csv file
    Lines under [Reads] are stored in a list with key name "readLengths"
    All other key names are translated according to the
        metadata_key_translation_dict

    arguments:
            sample_sheet_file -- path to UploadList.csv

    returns a dictionary containing the parsed key:pair values from .csv file
    """

    metadata_dict = {"readLengths": [], "indexCycles": []}

    csv_reader = common.get_csv_reader(sample_sheet_file)

    section = None

    for line in csv_reader:
        if "[Header]" in line or "[BCLConvert_Settings]" in line:
            section = "header"
            continue
        elif "[Reads]" in line:
            section = "reads"
            continue
        elif "[BCLConvert_Data]" in line:
            break
        elif line and line[0].startswith("["):
            section = "unknown"
            continue

        if not line or not line[0]:
            continue

        if not section:
            logging.debug("Sample sheet is missing important sections: no sections were found")
            raise exceptions.SampleSheetError("Sample sheet is missing important sections: no sections were found.",
                                              sample_sheet_file)
        elif section == "reads":
            if line[0] == "Read1Cycles" or line[0] == "Read2Cycles":
                metadata_dict["readLengths"].append(line[1])
            elif line[0] == "Index1Cycles" or line[0] == "Index2Cycles":
                metadata_dict["indexCycles"].append(line[1])

    # currently sends just the larger readLengths
    if len(metadata_dict["readLengths"]) == 2:
        metadata_dict["layoutType"] = "PAIRED_END"
    elif len(metadata_dict["readLengths"]) == 1:
        metadata_dict["layoutType"] = "SINGLE_END"
    else:
        logging.debug("The sample sheet has invalid [Reads] sections: [Reads] section should have 2 or 4 entries.")
        raise exceptions.SampleSheetError("The sample sheet has invalid [Reads] sections: "
                                          "[Reads] section should have 2 or 4 entries.",
                                          sample_sheet_file)
    metadata_dict["readLengths"] = max(metadata_dict["readLengths"])
    metadata_dict["indexCycles"] = max(metadata_dict["indexCycles"])

    return metadata_dict


def parse_sample_list(sample_sheet_file, run_data_directory, run_data_directory_file_list):
    """
    Creates a list of Sample Objects

    :param sample_sheet_file: Sample Sheet file
    :param run_data_directory: Data directory including run directory (e.g. my_run/Data/Intensities/BaseCalls)
    :param run_data_directory_file_list: The list of all files in the data directory
    :return: list of Sample objects
    """
    sample_list = _parse_samples(sample_sheet_file)

    for sample in sample_list:
        properties_dict = _parse_out_sequence_file(sample)
        # this is the Illumina-defined pattern for naming fastq files, from:
        # http://blog.basespace.illumina.com/2014/08/18/fastq-upload-in-now-available-in-basespace/
        file_pattern = "{sample_name}_S{sample_number}_L\\d{{3}}_R(\\d+)_\\S+\\.fastq.*$".format(
            sample_name=re.escape(sample.sample_name), sample_number=sample.sample_number)
        logging.info("Looking for files with pattern {}".format(file_pattern))
        regex = re.compile(file_pattern)
        pf_list = list(filter(regex.search, run_data_directory_file_list))
        if not pf_list:
            # OK. So we didn't find any files using the **correct** file name
            # definition according to Illumina. Let's try again with our deprecated
            # behaviour, where we didn't actually care about the sample number:
            file_pattern = "{sample_name}_S\\d+_L\\d{{3}}_R(\\d+)_\\S+\\.fastq.*$".format(
                sample_name=re.escape(sample.sample_name))
            logging.info("Looking for files with pattern {}".format(file_pattern))

            regex = re.compile(file_pattern)
            pf_list = list(filter(regex.search, run_data_directory_file_list))

            if not pf_list:
                # we **still** didn't find anything. It's pretty likely, then that
                # there aren't any fastq files in the directory that match what
                # the sample sheet says...
                raise exceptions.SequenceFileError(
                    ("The uploader was unable to find an files with a file name that ends with "
                     ".fastq.gz for the sample in your sample sheet with name {} in the directory {}. "
                     "This usually happens when the Illumina MiSeq Reporter tool "
                     "does not generate any FastQ data.").format(
                        sample.sample_name, run_data_directory))

        # List of files may be invalid if directory searching in has been modified by user
        if not _validate_pf_list(pf_list):
            raise exceptions.SequenceFileError(
                ("The following file list {} found in the directory {} is invalid. "
                 "Please verify the folder containing the sequence files matches the SampleSheet file").format(
                    pf_list, run_data_directory))

        # Add the dir to each file to create the full path
        for i in range(len(pf_list)):
            pf_list[i] = path.join(run_data_directory, pf_list[i])

        sq = model.SequenceFile(file_list=pf_list, properties_dict=properties_dict)
        sample.sequence_file = deepcopy(sq)

    return sample_list


def _validate_pf_list(file_list):
    """
    Checks if list of files is valid:

    arguments:
            file_list -- list of file names that are grouped together

    returns:
            True: 1 file in list
            True: 2 files in list, where one contains `R1` in the correct position and the other contains `R2`
            False: Number of files <1 or >2, or 2 files do not contain `R1`/`R2` correctly
    """
    if len(file_list) < 1:  # Invalid
        return False
    if len(file_list) > 2:  # We should never expect more than 2 files, a forward & backwards read
        return False
    elif len(file_list) == 1:  # single read, valid
        return True
    else:
        try:
            # check if one file is R1 and other is R2
            regex_filter = ".*_S\\d+_L\\d{3}_R(\\d+)_\\S+\\.fastq.*$"
            n1 = int(re.search(regex_filter, file_list[0]).group(1))
            n2 = int(re.search(regex_filter, file_list[1]).group(1))
            return (n1 != n2) and (n1 == 1 or n1 == 2) and (n2 == 1 or n2 == 2)
        except AttributeError:  # Attribute error means the file had invalid text in it
            return False


def _parse_samples(sample_sheet_file):

    """
    Parse all the lines under "[Data]" in .csv file
    Keys in sample_key_translation_dict have their values changed for
        uploading to REST API
    All other keys keep the same name that they have in .csv file

    arguments:
            sample_sheet_file -- path to UploadList.csv

    returns	a list containing Sample objects that have been created by a
        dictionary from the parsed out key:pair values from .csv file
    """

    logging.info("Reading data from sample sheet {}".format(sample_sheet_file))

    csv_reader = common.get_csv_reader(sample_sheet_file)
    # start with an ordered dictionary so that keys are ordered in the same
    # way that they are inserted.
    sample_dict = OrderedDict()
    sample_list = []

    sample_key_translation_dict = {
        'Sample_ID': 'sampleName',
        'Sample_Project': 'sample_project'
    }

    _parse_samples.sample_key_translation_dict = sample_key_translation_dict

    # initialize dictionary keys from first line (data headers/attributes)
    set_attributes = False
    for line in csv_reader:

        if set_attributes:
            for item in line:

                if item in sample_key_translation_dict:
                    key_name = sample_key_translation_dict[item]
                else:
                    key_name = item

                sample_dict[key_name] = ""

            break

        if "[BCLConvert_Data]" in line:
            set_attributes = True

    # fill in values for keys. line is currently below the [Data] headers
    for sample_number, line in enumerate(csv_reader):

        if len(sample_dict.keys()) != len(line):
            """
            if there is one more Data header compared to the length of
            data values then add an empty string to the end of data values
            i.e the Description will be empty string
            assumes the last Data header is going to be the Description
            this handles the case where the last trailing comma is trimmed

            Shaun said this issue may come up when a user edits the
            SampleSheet from within the MiSeq software
            """
            if len(sample_dict.keys()) - len(line) == 1:
                line.append("")
            else:
                raise exceptions.SampleSheetError(
                    ("Your sample sheet is malformed. Expected to find {} "
                     "columns in the [Data] section, but only found {} columns "
                     "for line {}.".format(len(sample_dict.keys()), len(line), line)),
                    sample_sheet_file
                )

        for index, key in enumerate(sample_dict.keys()):
            sample_dict[key] = line[index].strip()  # assumes values are never empty

        new_sample_dict = deepcopy(sample_dict)
        new_sample_name = new_sample_dict['sampleName']
        del new_sample_dict['sampleName']

        sample = model.Sample(
            sample_name=new_sample_name,
            description="",
            sample_number=sample_number + 1,
            samp_dict=new_sample_dict)
        sample_list.append(sample)

    return sample_list


def _parse_out_sequence_file(sample):

    """
    Removes keys in argument sample that are not in sample_keys and
        stores them in sequence_file_dict

    arguments:
            sample -- Sample object
            the dictionary inside the Sample object is changed

    returns a dictionary containing keys not in sample_keys to be used to
        create a SequenceFile object
    """

    sample_keys = ["sampleName", "sample_project"]
    sequence_file_dict = {}
    sample_dict = sample.get_uploadable_dict()
    for key in list(sample_dict.keys()):
        if key not in sample_keys:
            sequence_file_dict[key] = sample_dict[key]

    return sequence_file_dict



import unittest
from os import path

from iridauploader.tests_integration import tests_integration

import iridauploader.api as api
import iridauploader.model as model


path_to_module = path.dirname(__file__)
if len(path_to_module) == 0:
    path_to_module = '.'


class TestApiIntegration(unittest.TestCase):

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)
        self.test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

    def test_send_and_get_project(self):
        """
        Tests sending and receiving project data
        :return:
        """
        # try sending a project
        project_name = "test_project"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)

        json_res = self.test_api.send_project(project)

        # make sure the json response is what we expect it to be
        self.assertEqual(json_res['resource']['name'], project_name)
        self.assertEqual(json_res['resource']['projectDescription'], project_description)

        # get a list of all projects
        proj_list = self.test_api.get_projects()
        proj = proj_list[len(proj_list) - 1]

        # verify info matches what we uploaded
        self.assertEqual(type(proj), model.Project)
        self.assertEqual(proj.name, project_name)
        self.assertEqual(proj.description, project_description)

        # make sure identifiers match
        self.assertEqual(json_res['resource']['identifier'], str(len(proj_list)))
        self.assertEqual(json_res['resource']['identifier'], proj.id)

    def test_project_exists(self):
        """
        upload a project and check if project can be found with the projects_exists method
        :return:
        """
        project_name = "test_project_exists"
        project_description = "test_project_exists_description"
        project = model.Project(name=project_name, description=project_description)

        json_res = self.test_api.send_project(project)
        project_id = json_res['resource']['identifier']
        self.assertTrue(self.test_api.project_exists(project_id))

    def test_send_and_get_sample(self):
        """
        Tests sending and receiving sample data
        :return:
        """
        # set up a project to upload samples to
        project_name = "test_project_2"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)

        proj_json_res = self.test_api.send_project(project)
        project_identifier = proj_json_res['resource']['identifier']

        # upload a sample
        sample_name = "test_sample"
        sample_desc = "test_sample_desc"
        sample = model.Sample(sample_name, sample_desc)

        json_res = self.test_api.send_sample(sample, project_identifier)

        # make sure the returned values match what we tried to upload
        self.assertEqual(json_res['resource']['sampleName'], sample_name)
        self.assertEqual(json_res['resource']['description'], sample_desc)

        # fetch the sample from the server and confirm it again
        sample_id = int(json_res['resource']['identifier'])
        res_sample = self.test_api.get_sample_by_id(sample_id)
        self.assertEqual(res_sample.sample_name, sample_name)
        self.assertEqual(res_sample.description, sample_desc)

        # get a list of samples on our project and make sure they match what we uploaded
        sample_list = self.test_api.get_samples(project_identifier)

        self.assertEqual(len(sample_list), 1)
        self.assertEqual(type(sample_list[0]), model.Sample)
        self.assertEqual(sample_list[0].sample_name, sample_name)
        self.assertEqual(sample_list[0].description, sample_desc)

    def test_sample_exists(self):
        """
        Upload a sample and make sure it can be found with the sample_exists method
        :return:
        """
        # create a project to upload samples to
        project_name = "test_project_exists"
        project_description = "test_project_exists_description"
        project = model.Project(name=project_name, description=project_description)

        json_res = self.test_api.send_project(project)
        project_id = json_res['resource']['identifier']

        # create and upload a sample, and verify it exists
        sample_name = "test_sample_exists"
        sample_desc = "test_sample_exists_desc"
        sample = model.Sample(sample_name, sample_desc)

        self.test_api.send_sample(sample, project_id)
        self.assertNotEqual(self.test_api.get_sample_id(sample_name, project_id), False)

    def test_send_and_get_sequence_files(self):
        """
        Tests sending and receiving sequence files
        :return:
        """
        # upload a project
        project_name = "test_project_2"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)

        proj_json_res = self.test_api.send_project(project)
        project_identifier = proj_json_res['resource']['identifier']

        # upload a sample
        sample_name = "test_sample"
        sample_desc = "test_sample_desc"
        sample = model.Sample(sample_name, sample_desc)

        self.test_api.send_sample(sample, project_identifier)

        # upload sequence files
        sequence_file_list = [
            path.join(path_to_module, "fake_dir_data", "file_1.fastq.gz"),
            path.join(path_to_module, "fake_dir_data", "file_2.fastq.gz")
        ]
        sequence_file = model.SequenceFile(sequence_file_list)

        upload_id = self.test_api.create_seq_run({'layoutType': 'PAIRED_END'}, 'miseq')

        self.test_api.send_sequence_files(sequence_file, sample_name, project_identifier, upload_id)

        # verify sequence files match what we sent to IRIDA
        returned_sequence_files = self.test_api.get_sequence_files(project_identifier, sample_name)

        self.assertEqual(returned_sequence_files[0]['fileName'], 'file_1.fastq.gz')
        self.assertEqual(returned_sequence_files[1]['fileName'], 'file_2.fastq.gz')

    def test_create_and_get_sequencing_runs(self):
        """
        Create a new sequencing run and verify it can be gotten back from IRIDA
        This test tests paired end runs
        :return:
        """
        # create a new paired end sequencing run on IRIDA
        run_id = self.test_api.create_seq_run({'layoutType': 'PAIRED_END'}, 'miseq')

        # get a list of all sequencing runs and verify out new one is there and matches
        response = self.test_api.get_seq_runs()

        seq_run = None
        for run in response:
            if run['identifier'] == run_id:
                seq_run = run

        self.assertIsNotNone(seq_run)

        self.assertEqual(seq_run['layoutType'], 'PAIRED_END')
        self.assertEqual(seq_run['uploadStatus'], 'UPLOADING')

    def test_create_and_get_sequencing_runs_single_end(self):
        """
        Create a new sequencing run and verify it can be gotten back from IRIDA
        This one tests single end runs
        :return:
        """
        # create a new paired end sequencing run on IRIDA
        run_id = self.test_api.create_seq_run({'layoutType': 'SINGLE_END'}, 'miseq')

        # get a list of all sequencing runs and verify out new one is there and matches
        response = self.test_api.get_seq_runs()

        seq_run = None
        for run in response:
            if run['identifier'] == run_id:
                seq_run = run

        self.assertIsNotNone(seq_run)

        self.assertEqual(seq_run['layoutType'], 'SINGLE_END')
        self.assertEqual(seq_run['uploadStatus'], 'UPLOADING')

    def test_set_sequencing_runs(self):
        """
        Tests the various set sequencing run methods
        :return:
        """
        def get_seq_run(run_identifier):
            response = self.test_api.get_seq_runs()

            s_run = None
            for run in response:
                if run['identifier'] == run_identifier:
                    s_run = run

            return s_run

        run_id = self.test_api.create_seq_run({'layoutType': 'PAIRED_END'}, 'miseq')

        seq_run = get_seq_run(run_id)

        # Starts off as uploading
        self.assertEqual(seq_run['uploadStatus'], 'UPLOADING')

        # Test Error, uploading, and error
        self.test_api.set_seq_run_error(run_id)
        seq_run = get_seq_run(run_id)
        self.assertEqual(seq_run['uploadStatus'], 'ERROR')

        self.test_api.set_seq_run_uploading(run_id)
        seq_run = get_seq_run(run_id)
        self.assertEqual(seq_run['uploadStatus'], 'UPLOADING')

        self.test_api.set_seq_run_complete(run_id)
        seq_run = get_seq_run(run_id)
        self.assertEqual(seq_run['uploadStatus'], 'COMPLETE')



EXIT_CODE_ERROR = 1
EXIT_CODE_SUCCESS = 0


class ExitReturn:
    """
    When the program exits it returns this class.
    Contains an exit code (0 or 1)
    if exiting with error, an Exception object will be included

    This acts as a wrapper for exit status's so that exception info can be captured in gui, but cli can still return 0/1
    """

    def __init__(self, exit_code, error=None):
        self._exit_code = exit_code
        if type(error) is str:
            # If a string is given as the error, wrap in an exception.
            # todo: when reworking exceptions, this will likely be something that could be improved
            error = Exception(error)
        self._error = error

    @property
    def exit_code(self):
        return self._exit_code

    @property
    def error(self):
        return self._error



import unittest
from os import path
from unittest.mock import patch

import iridauploader.parsers.directory.sample_parser as sample_parser
from iridauploader.parsers.exceptions import SampleSheetError
import iridauploader.model as model

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestVerifySampleSheetFileNamesInFileList(unittest.TestCase):
    """
    test file existence verification function
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid(self):
        sheet_file = path.join(path_to_module, "fake_dir_data",
                               "SampleList_simple.csv")

        run_data_directory_file_list = ["file_1.fastq.gz", "file_2.fastq.gz"]

        sample_parser.verify_sample_sheet_file_names_in_file_list(sheet_file, run_data_directory_file_list)

    def test_file_names_do_not_match_paired_end(self):
        sheet_file = path.join(path_to_module, "fake_dir_data",
                               "SampleList_simple.csv")

        run_data_directory_file_list = ["file_1.fastq.gz", "file_a.fastq.gz"]

        with self.assertRaises(SampleSheetError):
            sample_parser.verify_sample_sheet_file_names_in_file_list(sheet_file, run_data_directory_file_list)

    def test_file_names_do_not_match_single_end(self):
        sheet_file = path.join(path_to_module, "fake_dir_data",
                               "no_reverse.csv")

        run_data_directory_file_list = ["file_a.fastq.gz"]

        with self.assertRaises(SampleSheetError):
            sample_parser.verify_sample_sheet_file_names_in_file_list(sheet_file, run_data_directory_file_list)


class TestBuildSampleListFromSampleSheetWithAbsPath(unittest.TestCase):
    """
    test parsing the list of samples from a sample sheet
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid(self):
        """
        Given a valid sample sheet, parse correctly
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_dir_data",
                               "SampleList_simple.csv")

        file_path_1 = path.join(path_to_module,
                                "fake_dir_data", "file_1.fastq.gz")
        file_path_2 = path.join(path_to_module,
                                "fake_dir_data", "file_2.fastq.gz")

        res = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(sheet_file)

        # Check we have 1 sample
        self.assertEqual(len(res), 1)
        # Check if data is correct
        self.assertEqual(res[0].sample_name, "my-sample-1")
        self.assertEqual(res[0].get_uploadable_dict()["sample_project"], "75")
        self.assertEqual(res[0].get_uploadable_dict()["File_Forward"], "file_1.fastq.gz")
        self.assertEqual(res[0].get_uploadable_dict()["File_Reverse"], "file_2.fastq.gz")
        self.assertEqual(res[0].sequence_file.file_list[0], file_path_1)
        self.assertEqual(res[0].sequence_file.file_list[1], file_path_2)

    @patch("iridauploader.parsers.directory.sample_parser._parse_samples")
    def test_valid_full_file_path(self, mock_parse_samples):
        """
        Given a valid sample sheet with full file paths, parse correctly
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_dir_data",
                               "SampleList_simple.csv")

        file_path_1 = path.join(path_to_module,
                                "fake_dir_data", "file_1.fastq.gz")
        file_path_2 = path.join(path_to_module,
                                "fake_dir_data", "file_2.fastq.gz")

        sample_list = [
            model.Sample(
                sample_name='my-sample-1',
                description="",
                sample_number=0,
                samp_dict={
                    ('sample_project', '75'),
                    ('File_Forward', path.abspath(file_path_1)),
                    ('File_Reverse', path.abspath(file_path_2))
                }
            )
        ]

        mock_parse_samples.return_value = sample_list

        res = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(sheet_file)

        mock_parse_samples.assert_called_with(sheet_file)
        # Check we have 1 sample
        self.assertEqual(len(res), 1)
        # Check if data is correct
        self.assertEqual(res[0].sample_name, "my-sample-1")
        self.assertEqual(res[0].get_uploadable_dict()["sample_project"], "75")
        self.assertEqual(res[0].get_uploadable_dict()["File_Forward"], path.abspath(file_path_1))
        self.assertEqual(res[0].get_uploadable_dict()["File_Reverse"], path.abspath(file_path_2))
        self.assertEqual(res[0].sequence_file.file_list[0], file_path_1)
        self.assertEqual(res[0].sequence_file.file_list[1], file_path_2)

    def test_no_forward_read(self):
        """
        No Valid files were found with names given in sample sheet
        :return:
        """
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "list_no_forward.csv")

        with self.assertRaises(SampleSheetError):
            res = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(file_path)

    def test_no_reverse_read(self):
        """
        The file list in the sample sheet is invalid
        :return:
        """
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "list_no_reverse.csv")

        with self.assertRaises(SampleSheetError):
            res = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(file_path)


class TestBuildSampleListFromSampleSheetNoVerify(unittest.TestCase):
    """
    test parsing the list of samples from a sample sheet
    These samplelists are built without verification that the files exist
    But we still check the positive test cases
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid(self):
        """
        Given a valid sample sheet, parse correctly
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_dir_data",
                               "SampleList_simple.csv")

        file_name_1 = "file_1.fastq.gz"
        file_name_2 = "file_2.fastq.gz"

        res = sample_parser.build_sample_list_from_sample_sheet_no_verify(sheet_file)

        # Check we have 1 sample
        self.assertEqual(len(res), 1)
        # Check if data is correct
        self.assertEqual(res[0].sample_name, "my-sample-1")
        self.assertEqual(res[0].get_uploadable_dict()["sample_project"], "75")
        self.assertEqual(res[0].get_uploadable_dict()["File_Forward"], file_name_1)
        self.assertEqual(res[0].get_uploadable_dict()["File_Reverse"], file_name_2)
        self.assertEqual(res[0].sequence_file.file_list[0], file_name_1)
        self.assertEqual(res[0].sequence_file.file_list[1], file_name_2)


class TestOnlySingleOrPairedInSampleList(unittest.TestCase):
    """
    test boolean of if there is only one of paired or single end files in a list of samples
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_mixed_paired_and_single_reads(self):
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "list_mixed.csv")

        sample_list = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(file_path)
        res = sample_parser.only_single_or_paired_in_sample_list(sample_list)

        self.assertFalse(res)

    def test_only_single_end_reads(self):
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "no_reverse.csv")

        sample_list = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(file_path)
        res = sample_parser.only_single_or_paired_in_sample_list(sample_list)

        self.assertTrue(res)

    def test_only_paired_end_reads(self):
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "SampleList.csv")

        sample_list = sample_parser.build_sample_list_from_sample_sheet_with_abs_path(file_path)
        res = sample_parser.only_single_or_paired_in_sample_list(sample_list)

        self.assertTrue(res)


class TestParseSamples(unittest.TestCase):
    """
    Test validity or invalidity of parsed samples
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid(self):
        """
        Given a valid sample sheet, parse correctly
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_dir_data",
                               "SampleList_simple.csv")

        file_name_1 = "file_1.fastq.gz"
        file_name_2 = "file_2.fastq.gz"

        res = sample_parser._parse_samples(sheet_file)

        # Check we have 1 sample
        self.assertEqual(len(res), 1)
        # Check if data is correct
        self.assertEqual(res[0].sample_name, "my-sample-1")
        self.assertEqual(res[0].get_uploadable_dict()["sample_project"], "75")
        self.assertEqual(res[0].get_uploadable_dict()["File_Forward"], file_name_1)
        self.assertEqual(res[0].get_uploadable_dict()["File_Reverse"], file_name_2)

    def test_no_forward_read(self):
        """
        No Valid files were found with names given in sample sheet
        :return:
        """
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "no_forward.csv")

        with self.assertRaises(SampleSheetError):
            res = sample_parser._parse_samples(file_path)

    def test_no_reverse_read(self):
        """
        The file list in the sample sheet is invalid
        :return:
        """
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "no_reverse.csv")

        res = sample_parser._parse_samples(file_path)
        # This should have an empty file reverse
        self.assertEqual(res[0]["File_Reverse"], "")

    def test_no_reverse_read_with_comma(self):
        """
        The file list in the sample sheet is invalid
        :return:
        """
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "no_reverse_with_comma.csv")

        res = sample_parser._parse_samples(file_path)
        # This should have an empty file reverse
        self.assertEqual(res[0]["File_Reverse"], "")

    def test_no_read_files_in_list(self):
        """
        The file list in the sample sheet is invalid
        :return:
        """
        directory = path.join(path_to_module, "fake_dir_data")
        file_path = path.join(directory, "no_read_files.csv")

        with self.assertRaises(SampleSheetError):
            res = sample_parser._parse_samples(file_path)



import logging
# PyQt needs to be imported like this because for whatever reason they decided not to include a __all__ = [...]
import PyQt5.QtCore as QtCore

from pprint import pformat

from iridauploader.core import upload, parsing_handler, exit_return, upload_helpers
from iridauploader.parsers import exceptions


class StatusThread(QtCore.QThread):
    """
    This class handles the arguments for uploading
    Is used an a separate thread to upload so that the GUI is still responsive while uploading
    """
    def __init__(self):
        super().__init__()
        self._directory = ""
        self._result = None

    def set_vars(self, directory):
        """
        Sets the variables in the object to the ones passed in
        :return:
        """
        self._directory = directory

    def get_result(self):
        """
        returns the result that was returned after run was called.
        :return: DirectoryStatus object
        """
        return self._result

    def run(self):
        """
        This runs when the threads start call is done
        :return: None
        """
        self._result = parsing_handler.get_run_status(self._directory)
        pass


class ParseThread(QtCore.QThread):
    """
    This class handles the arguments for uploading
    Is used an a separate thread to upload so that the GUI is still responsive while uploading
    """
    def __init__(self):
        super().__init__()
        self._directory = ""
        self._parse_as_partial = False
        self._run = None
        self._error = None

    def set_vars(self, directory, parse_as_partial):
        """
        Sets the variables in the object to the ones passed in
        :return:
        """
        self._directory = directory
        self._parse_as_partial = parse_as_partial

    def get_run(self):
        return self._run

    def get_error(self):
        return self._error

    def run(self):
        """
        This runs when the threads start call is done
        :return:
        """
        try:
            status = parsing_handler.get_run_status(self._directory)
            seq_run = parsing_handler.parse_and_validate(self._directory)
            if self._parse_as_partial:
                seq_run = upload_helpers.set_uploaded_samples_to_skip(seq_run, status.get_sample_status_list())
            self._run = seq_run

        except exceptions.DirectoryError as e:
            # Directory was not valid for some reason
            full_error = "GUI: ERROR! An error occurred with directory '{}', with message: {}".format(e.directory,
                                                                                                      e.message)
            logging.error(full_error)
            self._error = e.message
            self._run = None
        except exceptions.ValidationError as e:
            # Sequencing Run / SampleSheet was not valid for some reason
            error_msg = "GUI: ERROR! Errors occurred during validation with message: {}".format(e.message)
            logging.error(error_msg)
            error_list_msg = "GUI: Error list: " + pformat(e.validation_result.error_list)
            logging.error(error_list_msg)
            full_error = "Error: " + e.message + "\nError List:\n"
            for err in e.validation_result.error_list:
                full_error = full_error + str(err) + "\n"
            self._error = full_error
            self._run = None
        except exceptions.SampleSheetError as e:
            # SampleSheet was not valid for some reason
            error_msg = "GUI: ERROR! Errors occurred during parsing with message: {}".format(e.message)
            logging.error(error_msg)
            self._error = "Errors occurred during parsing `{}` with message: {}".format(e.errors, e.message)
            self._run = None
        except exceptions.SequenceFileError as e:
            # Sequence Files were invalid for some reason
            full_error = "GUI: ERROR! An error occurred while parsing: {}".format(e.message)
            logging.error(full_error)
            self._error = e.message
            self._run = None
        except Exception as e:
            # Some other error occurred
            full_error = "GUI: ERROR! An unknown error occurred while parsing: {}".format(str(e))
            logging.error(full_error)
            self._error = "ERROR! An error occurred while parsing: {}".format(str(e))
            self._run = None

        pass


class UploadThread(QtCore.QThread):
    """
    This class handles the arguments for uploading
    Is used an a separate thread to upload so that the GUI is still responsive while uploading
    """
    def __init__(self):
        super().__init__()
        self._run_dir = ""
        self._upload_mode = None
        self._exit_return = None
        self._partial_continue = False

    def set_vars(self, run_dir, upload_mode, partial_continue):
        """
        Sets the variables in the object to the ones passed in
        :param run_dir:
        :param upload_mode:
        :param partial_continue:
        :return:
        """
        self._run_dir = run_dir
        self._upload_mode = upload_mode
        self._partial_continue = partial_continue

    def run(self):
        """
        This runs when the threads start call is done

        When uploading, the default is to upload with force_upload=True,
          when continuing a partial upload, force_upload=False instead
          s.t. we ignore delay logic and simplify earlier run parse logic
        :return:
        """

        force_upload = not self._partial_continue
        self._exit_return = upload.upload_run_single_entry(directory=self._run_dir,
                                                           force_upload=force_upload,
                                                           upload_mode=self._upload_mode,
                                                           continue_upload=self._partial_continue)
        pass

    def is_success(self):
        if self._exit_return:
            return self._exit_return.exit_code == exit_return.EXIT_CODE_SUCCESS

    def get_exit_error(self):
        if self._exit_return:
            return self._exit_return.error



"""
Sample metadata can include:
samplePlate
sampleWell
i7IndexID
index
i5IndexID
index2
etc.
"""


class SequenceFile:

    uploadable_schema = {
        '_file_list': {
            'type': 'list',
            'empty': False,  # must have at least 1 file
            'nullable': False,
            'required': True,
            'schema': {'type': 'string'}
        },
        '_properties_dict': {'type': 'dict'}
    }

    def __init__(self, file_list, properties_dict=None):
        if properties_dict is None:
            self._properties_dict = {}
        else:
            self._properties_dict = properties_dict  # Sample metadata, needed run_id gets affixed in upload
        self._file_list = file_list

    @property
    def properties_dict(self):
        return self._properties_dict

    def get(self, key):
        ret_val = None
        if key in self._properties_dict:
            ret_val = self._properties_dict[key]
        return ret_val

    @property
    def file_list(self):
        return self._file_list

    def is_paired_end(self):
        return len(self._file_list) == 2

    def __str__(self):
        d = {'propertyDict': str(self._properties_dict), 'fileList': str(self._file_list)}
        return str(d)

    def get_dict(self):
        return self.__dict__



from iridauploader.api.exceptions.irida_connection_error import IridaConnectionError
from iridauploader.api.exceptions.irida_resource_error import IridaResourceError
from iridauploader.api.exceptions.irida_upload_canceled_exception import IridaUploadCanceledException
from iridauploader.api.exceptions.file_error import FileError



import unittest
from http import HTTPStatus
from unittest.mock import patch
from os import path

from iridauploader.core import api_handler

from iridauploader.parsers.miseq.parser import Parser
from iridauploader.api import MODE_DEFAULT, MODE_FAST5, MODE_ASSEMBLIES
from iridauploader.api.exceptions import IridaResourceError, IridaConnectionError, FileError
from iridauploader.model.exceptions import ModelValidationError
from iridauploader.model import DirectoryStatus

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestPrepareAndValidateForUpload(unittest.TestCase):
    """
    Tests the core.api_handler.prepare_and_validate_for_upload function
    """

    sequencing_run = None

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

        # make a sequencing run that we can work with in the tests
        sheet_file = path.join(path_to_module, "fake_ngs_data", "SampleSheet.csv")
        global sequencing_run
        sequencing_run = Parser().get_sequencing_run(sheet_file)

    def tearDown(self):
        # set out sequencing run back to None so each tests starts from a clean state
        global sequencing_run
        sequencing_run = None

    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_valid_all_functions_called(self, mock_api_instance):
        """
        Makes sure that all functions are called when a valid sequencing run in given
        :return:
        """
        global sequencing_run

        # create mocked data for each call that will occur
        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.project_exists.side_effect = [True]
        stub_api_instance.sample_exists.side_effect = [True, True, True]

        mock_api_instance.side_effect = [stub_api_instance]

        # check that each function was called the correct number of times and with the correct data
        res = api_handler.prepare_and_validate_for_upload(sequencing_run)
        stub_api_instance.project_exists.assert_called_once_with("6")
        stub_api_instance.sample_exists.assert_has_calls([
            unittest.mock.call('01-1111', '6'),
            unittest.mock.call('02-2222', '6'),
            unittest.mock.call('03-3333', '6')
        ])
        self.assertTrue(res.is_valid())

    def test_api_not_init(self):
        """
        Makes sure an exception is thrown when attempting to use api without initialization
        :return:
        """
        global sequencing_run

        # Api instance not init, so return None when fetched
        # mock_api_instance.side_effect = [None]

        # exception should be thrown
        with self.assertRaises(Exception):
            res = api_handler.prepare_and_validate_for_upload(sequencing_run)

    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_invalid_validation_project_does_not_exist(self, mock_api_instance):
        """
        Makes sure the returned validation result includes all types of invalid properties in the sequencing run
        Project does not exist on IRIDA
        :return:
        """
        global sequencing_run

        # create mock data for a non existant project
        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.project_exists.side_effect = [False]

        mock_api_instance.side_effect = [stub_api_instance]

        # verify the functions are called correctly
        res = api_handler.prepare_and_validate_for_upload(sequencing_run)
        stub_api_instance.project_exists.assert_called_once_with("6")

        # the result should be invalid with an IridaResourceError
        self.assertFalse(res.is_valid())
        self.assertEqual(res.error_count(), 1)
        self.assertEqual(type(res.error_list[0]), IridaResourceError)

    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_invalid_validation_user_does_not_have_project_permission(self, mock_api_instance):
        """
        Makes sure the returned validation result includes all types of invalid properties in the sequencing run
        User does not have project permission
        :return:
        """
        global sequencing_run

        # create mock data for a non existant project
        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.project_exists.side_effect = [False]
        stub_api_instance.try_project_access.side_effect = [HTTPStatus.UNAUTHORIZED]

        mock_api_instance.side_effect = [stub_api_instance]

        # verify the functions are called correctly
        res = api_handler.prepare_and_validate_for_upload(sequencing_run)
        stub_api_instance.project_exists.assert_called_once_with("6")

        # the result should be invalid with an IridaResourceError
        self.assertFalse(res.is_valid())
        self.assertEqual(res.error_count(), 1)
        self.assertEqual(type(res.error_list[0]), IridaResourceError)

    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_valid_send_sample(self, mock_api_instance):
        """
        Makes sure we try to send a sample to IRIDA if it doesn't exist
        :return:
        """
        global sequencing_run

        # create mocks for each function call that will occur
        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.project_exists.side_effect = [True]
        stub_api_instance.sample_exists.side_effect = [True, True, False, True]
        stub_api_instance.send_sample.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]

        res = api_handler.prepare_and_validate_for_upload(sequencing_run)

        # Ensure we check the existence of the sample after uploading it
        stub_api_instance.sample_exists.assert_has_calls([
            unittest.mock.call('01-1111', '6'),
            unittest.mock.call('02-2222', '6'),
            unittest.mock.call('03-3333', '6'),
            unittest.mock.call('03-3333', '6'),
        ])

        self.assertTrue(res.is_valid())

    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_invalid_could_not_send_sample(self, mock_api_instance):
        """
        Makes sure we try to send a sample to IRIDA if it doesn't exist,
            and IridaResourceError is thrown when it cannot send the sample
        :return:
        """
        global sequencing_run

        # create mocks for each function call that will occur
        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.project_exists.side_effect = [True]
        stub_api_instance.sample_exists.side_effect = [True, True, False, False]
        stub_api_instance.send_sample.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]

        res = api_handler.prepare_and_validate_for_upload(sequencing_run)

        # Ensure we check the existence of the sample after uploading it
        stub_api_instance.sample_exists.assert_has_calls([
            unittest.mock.call('01-1111', '6'),
            unittest.mock.call('02-2222', '6'),
            unittest.mock.call('03-3333', '6'),
            unittest.mock.call('03-3333', '6'),
        ])

        # check that it's invalid and the correct error is included
        self.assertFalse(res.is_valid())
        self.assertEqual(res.error_count(), 1)
        self.assertEqual(type(res.error_list[0]), IridaResourceError)

    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_invalid_connection_error(self, mock_api_instance):
        """
        Makes sure we try to send a sample to IRIDA if it doesn't exist,
            and IridaConnectionError is thrown when it cannot send the sample
        :return:
        """
        global sequencing_run

        # create mocks for each function call that will occur
        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.project_exists.side_effect = [True]
        stub_api_instance.sample_exists.side_effect = [True, True, False]
        stub_api_instance.send_sample.side_effect = [IridaConnectionError]

        mock_api_instance.side_effect = [stub_api_instance]

        res = api_handler.prepare_and_validate_for_upload(sequencing_run)

        # Ensure we check the existence of the sample after uploading it
        stub_api_instance.sample_exists.assert_has_calls([
            unittest.mock.call('01-1111', '6'),
            unittest.mock.call('02-2222', '6'),
            unittest.mock.call('03-3333', '6'),
        ])

        # check that it's invalid and the correct error is included
        self.assertFalse(res.is_valid())
        self.assertEqual(res.error_count(), 1)
        self.assertEqual(type(res.error_list[0]), IridaConnectionError)

    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_invalid_resource_error(self, mock_api_instance):
        """
        Makes sure we try to send a sample to IRIDA if it doesn't exist,
            and IridaResourceError is thrown when it cannot send the sample
        :return:
        """
        global sequencing_run

        # create mocks for each function call that will occur
        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.project_exists.side_effect = [True]
        stub_api_instance.sample_exists.side_effect = [True, True, False]
        stub_api_instance.send_sample.side_effect = [IridaResourceError("")]

        mock_api_instance.side_effect = [stub_api_instance]

        res = api_handler.prepare_and_validate_for_upload(sequencing_run)

        # Ensure we check the existence of the sample after uploading it
        stub_api_instance.sample_exists.assert_has_calls([
            unittest.mock.call('01-1111', '6'),
            unittest.mock.call('02-2222', '6'),
            unittest.mock.call('03-3333', '6'),
        ])

        # check that it's invalid and the correct error is included
        self.assertFalse(res.is_valid())
        self.assertEqual(res.error_count(), 1)
        self.assertEqual(type(res.error_list[0]), IridaResourceError)


class TestUploadSequencingRun(unittest.TestCase):
    """
    Tests the core.api_handler.upload_sequencing_run function
    """

    class StubDirectoryStatus:
        run_id = None
        status = None

        def set_sample_uploaded(self, sample_name, project_id, uploaded):
            return None

    sequencing_run = None

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)
        # make a sequencing run that we can work with in the tests
        sheet_file = path.join(path_to_module, "fake_ngs_data", "SampleSheet.csv")
        global sequencing_run
        sequencing_run = Parser().get_sequencing_run(sheet_file)

    def tearDown(self):
        # reset out make sequencing run back to None so all tests start from clean
        global sequencing_run
        sequencing_run = None

    @patch("iridauploader.core.api_handler._get_api_instance")
    @patch("iridauploader.progress.write_directory_status")
    def test_valid_all_functions_called(self, mock_progress, mock_api_instance):
        """
        Makes sure that all functions are called when a valid sequencing run in given
        :return:
        """
        global sequencing_run

        # set up all the mock data for a fake upload
        for samp in sequencing_run.project_list[0].sample_list:
            samp.sequence_file = "mock_sample"

        mock_sequence_run_id = 55

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.create_seq_run.side_effect = [mock_sequence_run_id]
        stub_api_instance.set_seq_run_uploading.side_effect = [True]
        stub_api_instance.send_sequence_files.side_effect = [True, True, True]
        stub_api_instance.set_seq_run_complete.side_effect = [True]
        stub_directory_status = self.StubDirectoryStatus()

        mock_api_instance.side_effect = [stub_api_instance]
        mock_progress.side_effect = [None, None, None, None, None, None, None, None, None, None]

        api_handler.upload_sequencing_run(sequencing_run,
                                          directory_status=stub_directory_status,
                                          upload_mode=MODE_DEFAULT)

        # ensure the response matches our mocks, and that all the needed functions were called correctly
        stub_api_instance.create_seq_run.assert_called_once_with(sequencing_run.metadata,
                                                                 sequencing_run.sequencing_run_type)
        stub_api_instance.set_seq_run_uploading.assert_called_once_with(mock_sequence_run_id)
        stub_api_instance.send_sequence_files.assert_has_calls([
            unittest.mock.call(project_id='6', sample_name='01-1111', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_DEFAULT),
            unittest.mock.call(project_id='6', sample_name='02-2222', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_DEFAULT),
            unittest.mock.call(project_id='6', sample_name='03-3333', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_DEFAULT)
        ])
        stub_api_instance.set_seq_run_complete.assert_called_once_with(mock_sequence_run_id)
        # Verify the DirectoryStatus object got assigned a run_id and status for upload
        self.assertEqual(stub_directory_status.status, DirectoryStatus.PARTIAL)
        self.assertEqual(stub_directory_status.run_id, mock_sequence_run_id)

    @patch("iridauploader.core.api_handler._get_api_instance")
    @patch("iridauploader.progress.write_directory_status")
    def test_valid_all_functions_called_assemblies(self, mock_progress, mock_api_instance):
        """
        Makes sure that all functions are called when a valid sequencing run in given
        :return:
        """
        global sequencing_run

        # set up all the mock data for a fake upload
        for samp in sequencing_run.project_list[0].sample_list:
            samp.sequence_file = "mock_sample"
        sequencing_run.assemblies = True

        mock_sequence_run_id = 55

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.create_seq_run.side_effect = [mock_sequence_run_id]
        stub_api_instance.set_seq_run_uploading.side_effect = [True]
        stub_api_instance.send_sequence_files.side_effect = [True, True, True]
        stub_api_instance.set_seq_run_complete.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]
        mock_progress.side_effect = [None, None, None, None, None, None, None, None, None, None]

        api_handler.upload_sequencing_run(sequencing_run,
                                          directory_status=self.StubDirectoryStatus(),
                                          upload_mode=MODE_ASSEMBLIES)

        # ensure the response matches our mocks, and that all the needed functions were called correctly
        stub_api_instance.create_seq_run.assert_called_once_with(sequencing_run.metadata,
                                                                 sequencing_run.sequencing_run_type)
        stub_api_instance.set_seq_run_uploading.assert_called_once_with(mock_sequence_run_id)
        stub_api_instance.send_sequence_files.assert_has_calls([
            unittest.mock.call(project_id='6', sample_name='01-1111', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_ASSEMBLIES),
            unittest.mock.call(project_id='6', sample_name='02-2222', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_ASSEMBLIES),
            unittest.mock.call(project_id='6', sample_name='03-3333', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_ASSEMBLIES)
        ])
        stub_api_instance.set_seq_run_complete.assert_called_once_with(mock_sequence_run_id)

    @patch("iridauploader.core.api_handler._get_api_instance")
    @patch("iridauploader.progress.write_directory_status")
    def test_valid_all_functions_called_fast5(self, mock_progress, mock_api_instance):
        """
        Makes sure that all functions are called when a valid sequencing run in given
        :return:
        """
        global sequencing_run

        # set up all the mock data for a fake upload
        for samp in sequencing_run.project_list[0].sample_list:
            samp.sequence_file = "mock_sample"
        sequencing_run.assemblies = True

        mock_sequence_run_id = 55

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.create_seq_run.side_effect = [mock_sequence_run_id]
        stub_api_instance.set_seq_run_uploading.side_effect = [True]
        stub_api_instance.send_sequence_files.side_effect = [True, True, True]
        stub_api_instance.set_seq_run_complete.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]
        mock_progress.side_effect = [None, None, None, None, None, None, None, None, None, None]

        api_handler.upload_sequencing_run(sequencing_run,
                                          directory_status=self.StubDirectoryStatus(),
                                          upload_mode=MODE_FAST5)

        # ensure the response matches our mocks, and that all the needed functions were called correctly
        stub_api_instance.create_seq_run.assert_called_once_with(sequencing_run.metadata,
                                                                 sequencing_run.sequencing_run_type)
        stub_api_instance.set_seq_run_uploading.assert_called_once_with(mock_sequence_run_id)
        stub_api_instance.send_sequence_files.assert_has_calls([
            unittest.mock.call(project_id='6', sample_name='01-1111', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_FAST5),
            unittest.mock.call(project_id='6', sample_name='02-2222', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_FAST5),
            unittest.mock.call(project_id='6', sample_name='03-3333', sequence_file='mock_sample',
                               upload_id=55, upload_mode=MODE_FAST5)
        ])
        stub_api_instance.set_seq_run_complete.assert_called_once_with(mock_sequence_run_id)

    @patch("iridauploader.core.api_handler._get_api_instance")
    @patch("iridauploader.progress.write_directory_status")
    def test_invalid_resource_error_raised(self, mock_progress, mock_api_instance):
        """
        Makes sure that the sequencing run is set to error when an exception is thrown
        :return:
        """
        global sequencing_run

        # create mock data for our invalid sequencing run
        for samp in sequencing_run.project_list[0].sample_list:
            samp.sequence_file = "mock_sample"

        mock_sequence_run_id = 55

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.create_seq_run.side_effect = [mock_sequence_run_id]
        stub_api_instance.set_seq_run_uploading.side_effect = IridaResourceError("Boom")
        stub_api_instance.set_seq_run_error.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]
        mock_progress.side_effect = [None]

        # make sure the IridaResourceError was thrown correctly
        with self.assertRaises(IridaResourceError):
            api_handler.upload_sequencing_run(sequencing_run,
                                              directory_status=self.StubDirectoryStatus(),
                                              upload_mode=MODE_DEFAULT)

        # verify the sequencing run was set to an error state after the upload was run
        stub_api_instance.create_seq_run.assert_called_once_with(sequencing_run.metadata,
                                                                 sequencing_run.sequencing_run_type)
        stub_api_instance.set_seq_run_uploading.assert_called_once_with(mock_sequence_run_id)
        stub_api_instance.set_seq_run_error.assert_called_once_with(mock_sequence_run_id)

    @patch("iridauploader.core.api_handler._get_api_instance")
    @patch("iridauploader.progress.write_directory_status")
    def test_invalid_connection_error_raised(self, mock_progress, mock_api_instance):
        """
        Makes sure that the sequencing run is set to error when an exception is thrown
        :return:
        """
        global sequencing_run

        # create mock data for our invalid sequencing run
        for samp in sequencing_run.project_list[0].sample_list:
            samp.sequence_file = "mock_sample"

        mock_sequence_run_id = 55

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.create_seq_run.side_effect = [mock_sequence_run_id]
        stub_api_instance.set_seq_run_uploading.side_effect = IridaConnectionError()
        stub_api_instance.set_seq_run_error.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]
        mock_progress.side_effect = [None]

        # make sure the IridaResourceError was thrown correctly
        with self.assertRaises(IridaConnectionError):
            api_handler.upload_sequencing_run(sequencing_run,
                                              directory_status=self.StubDirectoryStatus(),
                                              upload_mode=MODE_DEFAULT)

        # verify the sequencing run was set to an error state after the upload was run
        stub_api_instance.create_seq_run.assert_called_once_with(sequencing_run.metadata,
                                                                 sequencing_run.sequencing_run_type)
        stub_api_instance.set_seq_run_uploading.assert_called_once_with(mock_sequence_run_id)
        stub_api_instance.set_seq_run_error.assert_called_once_with(mock_sequence_run_id)

    @patch("iridauploader.core.api_handler._get_api_instance")
    @patch("iridauploader.progress.write_directory_status")
    def test_invalid_file_error_raised(self, mock_progress, mock_api_instance):
        """
        Makes sure that the sequencing run is set to error when an exception is thrown
        :return:
        """
        global sequencing_run

        # create mock data for our invalid sequencing run
        for samp in sequencing_run.project_list[0].sample_list:
            samp.sequence_file = "mock_sample"

        mock_sequence_run_id = 55

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.create_seq_run.side_effect = [mock_sequence_run_id]
        stub_api_instance.set_seq_run_uploading.side_effect = FileError()
        stub_api_instance.set_seq_run_error.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]
        mock_progress.side_effect = [None]

        # make sure the IridaResourceError was thrown correctly
        with self.assertRaises(FileError):
            api_handler.upload_sequencing_run(sequencing_run,
                                              directory_status=self.StubDirectoryStatus(),
                                              upload_mode=MODE_DEFAULT)

        # verify the sequencing run was set to an error state after the upload was run
        stub_api_instance.create_seq_run.assert_called_once_with(sequencing_run.metadata,
                                                                 sequencing_run.sequencing_run_type)
        stub_api_instance.set_seq_run_uploading.assert_called_once_with(mock_sequence_run_id)
        stub_api_instance.set_seq_run_error.assert_called_once_with(mock_sequence_run_id)


class TestSendProject(unittest.TestCase):
    """
    Tests the core.api_handler.test_send_project function
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.api_handler.model_validator")
    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_all_functions_called(self, mock_api_instance, mock_model_validator):
        """
        Makes sure that all relevant functions are called when sending a project
        :return:
        """

        # mock data for sending a project
        mock_project = "mock_project"

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.send_project.side_effect = [True]

        mock_api_instance.side_effect = [stub_api_instance]

        mock_model_validator.validate_send_project.side_effect = [True]

        # send the mock project
        api_handler.send_project(mock_project)

        # check that all the key functions are called with the mock data given
        stub_api_instance.send_project.assert_called_once_with(mock_project)
        mock_model_validator.validate_send_project.assert_called_once_with(mock_project)

    @patch("iridauploader.core.api_handler.model_validator")
    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_can_not_send_project(self, mock_api_instance, mock_model_validator):
        """
        Makes sure that an IridaResourceError is thrown when the api is unable to upload
            (IridaResourceError) when uploading a project
        :return:
        """

        mock_project = "mock_project"

        stub_api_instance = unittest.mock.MagicMock()
        stub_api_instance.send_project.side_effect = IridaResourceError("BOOM")

        mock_api_instance.side_effect = [stub_api_instance]

        mock_model_validator.validate_send_project.side_effect = [True]

        with self.assertRaises(IridaResourceError):
            api_handler.send_project(mock_project)

        mock_model_validator.validate_send_project.assert_called_once_with(mock_project)

    @patch("iridauploader.core.api_handler.model_validator")
    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_project_invalid(self, mock_api_instance, mock_model_validator):
        """
        Makes sure that an IridaResourceError is thrown when an invalid project is
            given (ModelValidationError) when uploading a project
        :return:
        """

        mock_project = "mock_project"

        stub_api_instance = unittest.mock.MagicMock()
        mock_api_instance.side_effect = [stub_api_instance]

        mock_model_validator.validate_send_project.side_effect = ModelValidationError("BOOM", None)

        with self.assertRaises(IridaResourceError):
            api_handler.send_project(mock_project)

    @patch("iridauploader.core.api_handler.model_validator")
    @patch("iridauploader.core.api_handler._get_api_instance")
    def test_invalid_connection_error(self, mock_api_instance, mock_model_validator):
        """
        Makes sure that an IridaConnectionError is thrown when an invalid project is
            given (ModelValidationError) when uploading a project
        :return:
        """

        mock_project = "mock_project"

        stub_api_instance = unittest.mock.MagicMock()
        mock_api_instance.side_effect = [stub_api_instance]

        mock_model_validator.validate_send_project.side_effect = [True]
        stub_api_instance.send_project.side_effect = IridaConnectionError()

        with self.assertRaises(IridaConnectionError):
            api_handler.send_project(mock_project)



import logging

from iridauploader.parsers import directory, miseq, miniseq, nextseq, nextseq2k_nml

supported_parsers = [
    'miseq',
    'miseq_v26',
    'miseq_v31',
    'miseq_win10_jun2021',
    'miniseq',
    'nextseq',
    'nextseq2k_nml',
    'iseq',
    'directory',
    'nanopore_assemblies',
    'seqfu',
]


def parser_factory(parser_type):
    """
    This factory creates and returns parser objects

    When creating a new parser, the parser type can be added here to enable it's usage.

    example:
        from parsers import parser_factory
        my_parser = parser_factory("directory")

    :param parser_type: a String of a valid parser name
    :return:
    """
    if parser_type in ['directory', 'nanopore_assemblies', 'seqfu']:
        logging.debug("Creating directory parser")
        return directory.Parser(parser_type_name=parser_type)
    if parser_type in ['miseq', 'miseq_v26']:
        logging.debug("Creating miseq (v26) parser")
        return miseq.Parser(parser_type_name=parser_type)
    if parser_type in ['miniseq', 'iseq', 'miseq_v31', 'miseq_win10_jun2021']:
        logging.debug("Creating miniseq/iseq/miseq_v31 parser")
        return miniseq.Parser(parser_type_name=parser_type)
    if parser_type == "nextseq":
        logging.debug("Creating nextseq parser")
        return nextseq.Parser(parser_type_name=parser_type)
    if parser_type == "nextseq_nml_classic":
        logging.debug("Creating nml custom nextseq parser")
        return nextseq.Parser(
            parser_type_name=parser_type,
            sample_sheet_override='SampleSheetClassic.csv'
        )
    if parser_type == "nextseq_nml_dual_upload":
        logging.debug("Creating nml custom nextseq parser with extra required files")
        return nextseq.Parser(
            parser_type_name=parser_type,
            additional_required_files=['RunMetadata.csv']
        )
    if parser_type == "nextseq_nml_strict_sample_name":
        logging.debug("Creating nml custom nextseq parser with strict sample name matching")
        return nextseq.Parser(
            parser_type_name=parser_type,
            strict_sample_name_matching=True
        )
    if parser_type == "nextseq2k_nml":
        logging.debug("Creating nextseq2k_nml parser")
        return nextseq2k_nml.Parser(parser_type_name=parser_type)
    raise AssertionError("Bad parser creation, invalid parser_type given: {}".format(parser_type))



import unittest
from unittest.mock import patch
from os import path
from csv import reader
from io import StringIO

from iridauploader import parsers
import iridauploader.parsers.nextseq2k_nml.sample_parser as sample_parser
from iridauploader.parsers.exceptions import SampleSheetError, SequenceFileError
import iridauploader.model as model

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestParseMetadata(unittest.TestCase):
    """
    Test building meta data from sample sheets with a csv reader
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_parse_metadata_paired_valid(self, mock_csv_reader):
        """
        When given a valid directory, ensure valid metadata is built
        paired end reads
        :return:
        """
        h_field_values = (
            "FileFormatVersion,2\n"
            "RunName,some_run\n"
            "InstrumentPlatform,NextSeq1k2k\n"
            "InstrumentType,NextSeq2000\n"
        )

        reads = (
            "Read1Cycles,101\n"
            "Read2Cycles,101\n"
            "Index1Cycles,8\n"
            "Index2Cycles,8\n"
        )

        d_headers = "Sample_ID,Index,Index2,Sample_Project"

        d_field_values = (
            "01A100001,AAAAAAAA,TTTTTTTT,6\n"
            "01A100002,GGGGGGGG,CCCCCCCC,6\n"
            "01A100003,GGGGGGGG,CCCCCCCC,6\n"
        )

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[BCLConvert_Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values, reads=reads, d_headers=d_headers, d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        metadata = sample_parser.parse_metadata(None)
        # The meta data we care about the most
        self.assertEqual(metadata['readLengths'], "101")
        self.assertEqual(metadata['layoutType'], "PAIRED_END")
        self.assertEqual(metadata['indexCycles'], "8")

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_parse_metadata_single_valid(self, mock_csv_reader):
        """
        When given a valid directory, ensure valid metadata is built
        single end reads
        :return:
        """
        h_field_values = (
            "FileFormatVersion,2\n"
            "RunName,some_run\n"
            "InstrumentPlatform,NextSeq1k2k\n"
            "InstrumentType,NextSeq2000\n"
        )

        reads = (
            "Read1Cycles,101\n"
            "Index1Cycles,8\n"
        )

        d_headers = "Sample_ID,Index,Index2,Sample_Project"

        d_field_values = (
            "01A100001,AAAAAAAA,TTTTTTTT,6\n"
            "01A100002,GGGGGGGG,CCCCCCCC,6\n"
            "01A100003,GGGGGGGG,CCCCCCCC,6\n"
        )

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[BCLConvert_Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values, reads=reads, d_headers=d_headers, d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        metadata = sample_parser.parse_metadata(None)
        self.assertEqual(metadata['layoutType'], "SINGLE_END")

    def test_parse_metadata(self):
        """
        Testing the parsing meta data with actual files, instead of mocked files
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "UploadList.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)

        correct_metadata = {"readLengths": "101",
                            "indexCycles": "8",
                            "layoutType": "PAIRED_END"}

        self.assertEqual(correct_metadata, meta_data)

    def test_parse_metadata_extra_commas(self):
        """
        Tests parsing meta data when extra commas are in the file
        Testing the parsing meta data with actual files, instead of mocked files
        :return:
        """
        sheet_file = path.join(path_to_module, "testValidSheetTrailingCommas",
                               "UploadList.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)

        correct_metadata = {"readLengths": "101",
                            "indexCycles": "8",
                            "layoutType": "PAIRED_END"}

        self.assertEqual(correct_metadata, meta_data)


class TestGetCsvReader(unittest.TestCase):
    """
    Test that the csv reader behaves as expected
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_get_csv_reader_valid_sheet(self):
        """
        Given a valid sample sheet, ensure the parsed sheet matches expected output
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "UploadList.csv")

        lines = parsers.common.get_csv_reader(sheet_file)

        correct_lines = [
            ['[Header]'],
            ['FileFormatVersion', '2'],
            ['RunName', 'some_run'],
            ['InstrumentPlatform', 'NextSeq1k2k'],
            ['InstrumentType', 'NextSeq2000'],
            [],
            ['[Reads]'],
            ['Read1Cycles', '101'],
            ['Read2Cycles', '101'],
            ['Index1Cycles', '8'],
            ['Index2Cycles', '8'],
            [],
            ['[BCLConvert_Settings]'],
            ['SoftwareVersion', '3.5.8'],
            ['AdapterRead1', 'CTGTCTCTTTTTTTTTTT'],
            ['AdapterRead2', 'CTGTCTCTCCCCCCCCCC'],
            [],
            ['[BCLConvert_Data]'],
            ['Sample_ID', 'Index', 'Index2', 'Sample_Project'],
            ['01A100001', 'AAAAAAAA', 'TTTTTTTT', '6'],
            ['01A100002', 'GGGGGGGG', 'CCCCCCCC', '6'],
            ['01A100003', 'GGGGGGGG', 'CCCCCCCC', '6']
        ]

        for line, c_line in zip(lines, correct_lines):
            self.assertEqual(line, c_line)

    def test_get_csv_reader_no_sheet(self):
        """
        Make sure an error is raised if the csv reader is not given a valid sample sheet
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "Data")

        with self.assertRaises(SampleSheetError):
            parsers.common.get_csv_reader(sheet_file)


class TestValidatePfList(unittest.TestCase):
    """
    Testing the various cases that can appear when validating the pf list
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid_lists(self):
        self.assertTrue(sample_parser._validate_pf_list(
            ['01A100001_S1_L001_R1_001.fastq.gz', '01-1111_S1_L001_R2_001.fastq.gz']
        ))
        self.assertTrue(sample_parser._validate_pf_list(
            ['01A100001_S1_L001_R1_001.fastq.gz']
        ))

    def test_invalid_lists(self):
        self.assertFalse(sample_parser._validate_pf_list(
            ['01A100001_S1_L001_R1_001.fastq.gz', '01-1111_S1_L001_R2_001.fastq.gz', '01-1111_S1_L001_R3_001.fastq.gz']
        ))

        self.assertFalse(sample_parser._validate_pf_list([]))

    def test_invalid_file_names(self):
        self.assertFalse(sample_parser._validate_pf_list(
            ['01A100001_S1_L001_R_001.fastq.gz', '01-1111_S1_L001_R_001.fastq.gz']
        ))
        self.assertFalse(sample_parser._validate_pf_list(
            ['01A100001_S1_L001_1_001.fastq.gz', '01-1111_S1_L001_2_001.fastq.gz']
        ))


class TestParseSampleList(unittest.TestCase):
    """
    test parsing the list of samples from a sample sheet
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid(self):
        """
        Ensure a a parsed valid directory matches the expected sample list
        :return:
        """
        directory = path.join(path_to_module, "fake_ngs_data")
        sheet_file = path.join(directory, "UploadList.csv")
        data_dir = path.join(directory, parsers.nextseq2k_nml.Parser.get_relative_data_directory())
        file_list = parsers.common.get_file_list(data_dir)

        sample = model.Sample(
            "01A100001",
            "",
            1,
            {
                "Index": "AAAAAAAA",
                "sample_project": "6",
                "Index2": "TTTTTTTT",
            }
        )

        sequence_file_properties = {
            'Index': 'AAAAAAAA',
            'Index2': 'TTTTTTTT',
            'description': "",
        }

        file_path_1 = path.join(path_to_module,
                                "fake_ngs_data", "Analysis", "1", "Data", "fast1", "01A100001_S1_L001_R1_001.fastq.gz")
        file_path_2 = path.join(path_to_module,
                                "fake_ngs_data", "Analysis", "1", "Data", "fast1", "01A100001_S1_L001_R2_001.fastq.gz")
        raw_file_list = [file_path_1, file_path_2]

        res = sample_parser.parse_sample_list(sample_sheet_file=sheet_file, run_data_directory=data_dir,
                                              run_data_directory_file_list=file_list)

        # Check sample is the same
        self.assertEqual(res[0].get_uploadable_dict(), sample.get_uploadable_dict())
        # Check sequencing file is correct
        self.assertEqual(res[0].sequence_file.properties_dict, sequence_file_properties)
        self.assertEqual(res[0].sequence_file.file_list.sort(), raw_file_list.sort())

    def test_not_pf_list(self):
        """
        No Valid files were found with names given in sample sheet
        :return:
        """
        directory = path.join(path_to_module, "ngs_not_pf_list")
        data_dir = path.join(directory, parsers.nextseq2k_nml.Parser.get_relative_data_directory())
        file_list = parsers.common.get_file_list(data_dir)
        file_path = path.join(directory, "UploadList.csv")

        with self.assertRaises(SequenceFileError):
            res = sample_parser.parse_sample_list(sample_sheet_file=file_path, run_data_directory=data_dir,
                                                  run_data_directory_file_list=file_list)

    def test_not_valid_pf_list(self):
        """
        The file list in the sample sheet is invalid
        :return:
        """
        directory = path.join(path_to_module, "ngs_not_valid_pf_list")
        data_dir = path.join(directory, parsers.nextseq2k_nml.Parser.get_relative_data_directory())
        file_list = parsers.common.get_file_list(data_dir)
        file_path = path.join(directory, "UploadList.csv")

        with self.assertRaises(SequenceFileError):
            res = sample_parser.parse_sample_list(sample_sheet_file=file_path, run_data_directory=data_dir,
                                                  run_data_directory_file_list=file_list)

    def test_space_in_sample_name(self):
        directory = path.join(path_to_module, "ngs_space_in_sample_name")
        data_dir = path.join(directory, parsers.nextseq2k_nml.Parser.get_relative_data_directory())
        file_list = parsers.common.get_file_list(data_dir)
        file_path = path.join(directory, "UploadList.csv")

        # Just making sure this doesn't throw an error
        sample_parser.parse_sample_list(sample_sheet_file=file_path, run_data_directory=data_dir,
                                        run_data_directory_file_list=file_list)


class TestParseSamples(unittest.TestCase):
    """
    Test validity or invalidity of parsed samples
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_parse_samples_valid(self):
        """
        Ensure a a parsed valid directory matches the expected samples
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "UploadList.csv")

        sample1 = model.Sample(
            "01A100001",
            "",
            1,
            {
                "Index": "AAAAAAAA",
                "sample_project": "6",
                "Index2": "TTTTTTTT",
                "description": ''
            }
        )

        sample2 = model.Sample(
            "01A100002",
            "",
            2,
            {
                "Index": "GGGGGGGG",
                "sample_project": "6",
                "Index2": "CCCCCCCC",
                "description": ''
            }
        )

        sample3 = model.Sample(
            "01A100003",
            "",
            3,
            {
                "Index": "GGGGGGGG",
                "sample_project": "6",
                "Index2": "CCCCCCCC",
                "description": ''
            }
        )

        correct_samples = [sample1, sample2, sample3]

        res = sample_parser._parse_samples(sheet_file)
        for r_sample, c_sample in zip(res, correct_samples):
            self.assertEqual(r_sample.get_uploadable_dict(), c_sample.get_uploadable_dict())


class TestParseOutSequenceFile(unittest.TestCase):
    """
    test building the sample object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_parse_out_sequence_file(self):
        """
        Tests that parse out sequence file correctly filters sample related data from the extra params dict
        And ensures that the uploadable dict correctly includes all the needed data after removal
        :return:
        """
        sample = model.Sample(
            "01A100003",
            "",
            3,
            {
                "Index": "GGGGGGGG",
                "sampleProject": "6",
                "Index2": "CCCCCCCC",
                "description": ''
            }
        )

        uploadable_dict = {'Index': 'GGGGGGGG',
                           'sampleName': '01A100003',
                           'sampleProject': '6',
                           'Index2': 'CCCCCCCC',
                           'description': ''}

        sequence_file_dict = {'Index': 'GGGGGGGG',
                              'sampleProject': '6',
                              'Index2': 'CCCCCCCC',
                              'description': ''}

        res = sample_parser._parse_out_sequence_file(sample)

        self.assertEqual(sample.get_uploadable_dict(), uploadable_dict)
        self.assertEqual(res, sequence_file_dict)



from iridauploader.model.sequencing_run import SequencingRun
from iridauploader.model.metadata import Metadata
from iridauploader.model.project import Project
from iridauploader.model.sample import Sample
from iridauploader.model.sequence_file import SequenceFile
from iridauploader.model.directory_status import DirectoryStatus
from iridauploader.model.validation_result import ValidationResult
from iridauploader.model import exceptions



import unittest

from iridauploader.api import api_calls


class TestIsIridaVersionCompatible(unittest.TestCase):
    """
    Tests the api.api_calls._is_irida_version_compatible function
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_simple_true(self):
        """
        Simple case
        :return:
        """
        result = api_calls.ApiCalls._is_irida_version_compatible(
            irida_version="23.01",
            minimum_irida_version="22.01"
        )
        self.assertTrue(result)

    def test_simple_false(self):
        """
        Simple case
        :return:
        """
        result = api_calls.ApiCalls._is_irida_version_compatible(
            irida_version="22.01",
            minimum_irida_version="23.01"
        )
        self.assertFalse(result)

    def test_simple_equal_true(self):
        """
        Simple case
        :return:
        """
        result = api_calls.ApiCalls._is_irida_version_compatible(
            irida_version="23.01",
            minimum_irida_version="23.01"
        )
        self.assertTrue(result)

    def test_unequal_length_true(self):
        """
        Simple case
        :return:
        """
        result = api_calls.ApiCalls._is_irida_version_compatible(
            irida_version="23.01.3",
            minimum_irida_version="22.01"
        )
        self.assertTrue(result)

    def test_unequal_length_false(self):
        """
        Simple case
        :return:
        """
        result = api_calls.ApiCalls._is_irida_version_compatible(
            irida_version="22.01.3",
            minimum_irida_version="23.01"
        )
        self.assertFalse(result)

    def test_simple_almost_equal_true(self):
        """
        Simple case
        :return:
        """
        result = api_calls.ApiCalls._is_irida_version_compatible(
            irida_version="23.01.2",
            minimum_irida_version="23.01.1"
        )
        self.assertTrue(result)

    def test_simple_almost_equal_false(self):
        """
        Simple case
        :return:
        """
        result = api_calls.ApiCalls._is_irida_version_compatible(
            irida_version="23.01.1",
            minimum_irida_version="23.01.2"
        )
        self.assertFalse(result)



"""
This file has api related core functionality
It handles initializing and managing an api instance as well as preparing IRIDA for an upload and starting the upload
"""

from http import HTTPStatus
import logging
import concurrent.futures

import iridauploader.api as api
import iridauploader.config as config
import iridauploader.model as model
import iridauploader.progress as progress
from iridauploader.core import model_validator

# The api instance is a global variable which lets the api behave like a singleton
# managed within this file
_api_instance = None


def _initialize_api(
        client_id, client_secret, base_url, username, password, timeout_multiplier, max_wait_time=20,
        http_max_retries=5, http_backoff_factor=0):
    """
    Creates the ApiCalls object from the api layer.
    Sets the instance to use the global _api_instance variable so it behaves as a singleton that can be easily re-init

    :param client_id:
    :param client_secret:
    :param base_url:
    :param username:
    :param password:
    :param timeout_multiplier:
    :param max_wait_time:
    :return: The ApiCalls instance
    """
    global _api_instance

    if not base_url.endswith('/api/'):
        logging.warning("base_url does not end in /api/, this configuration might be incorrect")

    _api_instance = api.ApiCalls(
        client_id=client_id,
        client_secret=client_secret,
        base_url=base_url,
        username=username,
        password=password,
        timeout_multiplier=timeout_multiplier,
        max_wait_time=max_wait_time,
        http_max_retries=http_max_retries,
        http_backoff_factor=http_backoff_factor,
    )
    return _api_instance


def _get_api_instance():
    """
    Returns the current ApiCalls instance, This function should be used anytime the ApiCalls object is needed

    Raises an exception if the api has not been initialized

    :return: the current ApiCalls instance
    """
    global _api_instance
    if _api_instance is None:
        logging.debug("_get_api_instance: The api instance has been called before being initialized.")
        raise Exception("The API has not been initialized")
    return _api_instance


def initialize_api_from_config():
    """
    Loads the api parameters from the config file and initializes the api with them

    :return: the api instance
    """
    client_id = config.read_config_option("client_id")
    client_secret = config.read_config_option("client_secret")
    base_url = config.read_config_option("base_url")
    username = config.read_config_option("username")
    password = config.read_config_option("password")
    timeout = config.read_config_option("timeout", expected_type=int)
    http_max_retries = config.read_config_option("http_max_retries", expected_type=int)
    http_backoff_factor = config.read_config_option("http_backoff_factor", expected_type=float)

    return _initialize_api(client_id=client_id,
                           client_secret=client_secret,
                           base_url=base_url,
                           username=username,
                           password=password,
                           timeout_multiplier=timeout,
                           http_max_retries=http_max_retries,
                           http_backoff_factor=http_backoff_factor,
                           )


def prepare_and_validate_for_upload(sequencing_run):
    """
    Prepares IRIDA to accept the sequencing run
    Validates that projects exist,
    Creates Samples on Projects on Irida if they do not exist yet

    Collects all errors during prep/validation in ValidationResult

    :param sequencing_run: SequencingRun object
    :return: ValidationResult object with all errors that raised while prepping
    """
    # get api
    api_instance = _get_api_instance()

    validation_result = model.ValidationResult()
    # Start online validation
    logging.debug("Checking existence of projects")
    for project in sequencing_run.project_list:
        # Validate project existence
        logging.debug("Checking existence of project: {}".format(project.id))
        if not api_instance.project_exists(project.id):
            if api_instance.try_project_access(project.id) in [HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN]:
                # Project exists but is not accessible by the user
                logging.debug("Project found but user does not have permission")
                err = api.exceptions.IridaResourceError("You do not have permission to access this project", project.id)
            else:
                # No project, add error to validation result and continue
                logging.debug("Could not find project: {}".format(project.id))
                err = api.exceptions.IridaResourceError("Project does not exist", project.id)
            validation_result.add_error(err)
            continue
        logging.debug("Project {} exists".format(project.id))

        # Validate sample existence
        logging.debug("Checking existence of samples")
        samples_to_create = []
        for sample in project.sample_list:
            logging.debug("Checking existence of Sample {} on Project {}".format(sample.sample_name, project.id))
            if api_instance.sample_exists(sample.sample_name, project.id):
                logging.debug("Sample {} exists on Project {}".format(sample.sample_name, project.id))
            else:
                samples_to_create.append((sample, project.id))

        # Create samples that do not exist
        samples_to_validate = []
        logging.debug("Creating {} samples".format(len(samples_to_create)))
        for sample_project_id_tuple in samples_to_create:
            sample = sample_project_id_tuple[0]
            project_id = sample_project_id_tuple[1]
            logging.debug("Sample {} was not found on project {}, creating new Sample".format(
                sample.sample_name, project_id))
            try:
                api_instance.send_sample(sample, project_id)
                # we only want tovalidate created samples, not ones that errored out
                samples_to_validate.append(sample_project_id_tuple)
                logging.debug("Sample Created")
            except (api.exceptions.IridaResourceError, api.exceptions.IridaConnectionError) as e:
                logging.debug("Sample could not be created")
                validation_result.add_error(e)
                continue

        # validate created samples were actually created
        logging.debug("validating creation of {} samples".format(len(samples_to_create)))
        for sample_project_id_tuple in samples_to_validate:
            sample = sample_project_id_tuple[0]
            project_id = sample_project_id_tuple[1]
            logging.debug("Verifying sample {} on project {} was created".format(sample.sample_name, project_id))
            if not api_instance.sample_exists(sample.sample_name, project_id):
                logging.debug("Sample was not created")
                err = api.exceptions.IridaResourceError("Could not create new Sample on Project {}", project_id)
                validation_result.add_error(err)
                continue

    return validation_result


def upload_sequencing_run(sequencing_run, directory_status, upload_mode, run_id=None):
    """
    Handles uploading a sequencing run

    Expects api to have been set up
    Expects sequencing run to have been validated
    Expects sequencing run to be valid for upload
    Expects directory_status to be valid

    :param sequencing_run: run to upload
    :param directory_status: DirectoryStatus object to update as files get uploaded
    :param upload_mode: mode of upload
    :param run_id: Default None, when given, run_id will be used instead of generating a new run_id
    :return:
    """
    # get api
    api_instance = _get_api_instance()

    # create a new seq run identifier if none is given
    if run_id is None:
        run_id = api_instance.create_seq_run(sequencing_run.metadata, sequencing_run.sequencing_run_type)
        logging.info("Sequencing run id '{}' has been created for this upload.".format(run_id))
    else:
        logging.info("Using existing run id '{}' for this upload.".format(run_id))
    # Update directory status file
    directory_status.run_id = run_id
    directory_status.status = model.DirectoryStatus.PARTIAL
    progress.write_directory_status(directory_status)

    try:
        # set seq run to upload
        api_instance.set_seq_run_uploading(run_id)
        # loop through projects
        for project in sequencing_run.project_list:
            # loop through samples
            for sample in project.sample_list:
                if sample.skip:
                    logging.info("Skipping Sample {} on Project {}, already uploaded."
                                 "".format(sample.sample_name, project.id))
                else:
                    logging.info("Uploading to Sample {} on Project {}".format(sample.sample_name, project.id))
                    # upload files
                    api_instance.send_sequence_files(sequence_file=sample.sequence_file,
                                                     sample_name=sample.sample_name,
                                                     project_id=project.id,
                                                     upload_id=run_id,
                                                     upload_mode=upload_mode)
                # Update status file on progress
                # Skipped samples are set to uploaded too, s.t. if they are skipped,
                #   and the upload fails and is continued again, they will be skipped again.
                directory_status.set_sample_uploaded(sample_name=sample.sample_name,
                                                     project_id=project.id,
                                                     uploaded=True)
                progress.write_directory_status(directory_status)

        # set seq run to complete
        api_instance.set_seq_run_complete(run_id)

        # set seq run to error if there is an error
    except api.exceptions.IridaConnectionError as e:
        logging.error("Failed to upload SequencingRun, Could not connect to IRIDA")
        api_instance.set_seq_run_error(run_id)
        raise e
    except api.exceptions.IridaResourceError as e:
        logging.error("Failed to upload SequencingRun, Could not access resources on IRIDA")
        api_instance.set_seq_run_error(run_id)
        raise e
    except api.exceptions.FileError as e:
        logging.error("Failed to upload SequencingRun, Could not access files to upload to IRIDA")
        api_instance.set_seq_run_error(run_id)
        raise e


def send_project(project):
    """
    Validates and sends a project object to IRIDA

    :param project: Project object
    :return: None
    """
    # get api
    api_instance = _get_api_instance()

    # validate project object
    try:
        model_validator.validate_send_project(project)
    except model.exceptions.ModelValidationError:
        raise api.exceptions.IridaResourceError("Project is invalid: check your project name and description")

    try:
        api_instance.send_project(project)
    except api.exceptions.IridaResourceError as e:
        logging.error("Failed to send project to IRIDA")
        raise e
    except api.exceptions.IridaConnectionError as e:
        logging.error("Failed to send project to IRIDA")
        raise e


def get_upload_modes():
    """
    Returns all the upload modes the api supports for sequence files

    :return: List of Strings
    """
    return api.UPLOAD_MODES


def get_default_upload_mode():
    """
    Returns the string for the default upload mode

    :return: Default upload mode string
    """
    return api.MODE_DEFAULT



def send_message(message):
    """
    This will eventually handle the message passing from the api layer
    Right now it is just a dummy function that prints
    """
    print(message)
    return


"""
Below is work in progress (non working) code of how the code could flow
to be redone once the UI is implemented
"""

#
# class MessageTopics(object):
#     upload_progress_topic = "finished_run_scan"
#     upload_started_topic = "run_discovered"
#     upload_completed_topic = "garbled_sample_sheet"
#     upload_failed_topic = "missing_files"
#
#
# class ApiMessage(object):
#     # todo: add run_id to this too?
#     def __init__(self,
#                  topic,
#                  exception=None,
#                  project_id=None,
#                  sample_name=None,
#                  sequence_file_name=None,
#                  progress=None):
#         self._topic = topic
#         self._exception = exception
#         self._project_id = project_id
#         self._sample_name = sample_name
#         self._sequence_file_name = sequence_file_name
#         self._progress = progress
#
#     @property
#     def property(self):
#         return self._topic
#
#     @property
#     def exception(self):
#         return self._exception
#
#     @property
#     def project_id(self):
#         return self._project_id
#
#     @property
#     def sample_name(self):
#         return self._sample_name
#
#     @property
#     def sequence_file_name(self):
#         return self._sequence_file_name
#
#     @property
#     def progress(self):
#         return self._progress
#
#     def __str__(self):
#         return "topic: " + self._topic + \
#                ", project_id: " + self.project_id + \
#                ", sample_name: " + self._sample_name + \
#                ", sequence_file_name: " + self._sequence_file_name



from cerberus import Validator

import iridauploader.model as model


def validate_sequencing_run(sequencing_run):
    """
    Validate a SequencingRun object for upload to irida

    If the parser is working as intended (and has proper validation) this should never return
    a ValidationResult with a list of errors. This function should be used when building a model
    from scratch, building a new parser, and as a final redundancy on parsers.

    :param sequencing_run: SequencingRun object to validate
    :return: ValidationResult object with list of errors if any
    """
    validation_result = model.ValidationResult()

    # Validation objects
    v_sequencing_run = Validator(model.SequencingRun.uploadable_schema, allow_unknown=True)
    v_project = Validator(model.Project.uploadable_schema, allow_unknown=True)
    v_sample = Validator(model.Sample.uploadable_schema, allow_unknown=True)
    v_sequence_file = Validator(model.SequenceFile.uploadable_schema, allow_unknown=True)

    # validation is nested so we can catch multiple levels of project/sample/file errors

    # Validate base SequencingRun Object
    try:
        _validate_object(v_sequencing_run, sequencing_run)

        # Validate projects in sequencing run
        for p in sequencing_run.project_list:
            try:
                _validate_object(v_project, p)

                # Validate samples in project
                for s in p.sample_list:
                    try:
                        _validate_object(v_sample, s)

                        # Validate SequenceFile on Sample
                        _validate_object(v_sequence_file, s.sequence_file)

                        # Validate tricky sequence_file rule
                        _validate_sequence_file_names(s.sequence_file)

                    except model.exceptions.ModelValidationError as e:
                        validation_result.add_error(e)

            except model.exceptions.ModelValidationError as e:
                validation_result.add_error(e)

    except model.exceptions.ModelValidationError as e:
        validation_result.add_error(e)

    return validation_result


def validate_send_project(project):
    """
    Validates a project object when using the send_project api call

    :return: raises a ModelValidationError when project is invalid
    """

    v_project = Validator(model.Project.send_project_schema, allow_unknown=True)
    _validate_object(v_project, project)


def _validate_object(validator, o):
    if not validator.validate(o.get_dict()):
        raise model.exceptions.ModelValidationError(validator.errors, o)


def _validate_sequence_file_names(sequence_file):
    """
    Validates that sequence files that are paired end have forward/reverse identifiers in the file names

    On the first difference found between the file names, we check to see that one file has one of [F, f, 1]
    and the other file has one of [R, r, 2]

    :param sequence_file:
    :return: raises a ModelValidationError when sequence_file is invalid
    """
    # Only applies to paired end files
    if not sequence_file.is_paired_end():
        return

    forward = ['F', 'f', '1']
    reverse = ['R', 'r', '2']

    file_list = sequence_file.file_list
    for letter_a, letter_b in zip(file_list[0], file_list[1]):
        if letter_a != letter_b:
            if ((letter_a in forward) and (letter_b in reverse)) or\
               ((letter_b in forward) and (letter_a in reverse)):
                return
            else:
                raise model.exceptions.ModelValidationError("Sequence File names are invalid. "
                                                            "First letter that is different between files should "
                                                            "identify forward/reverse with one of ['F', 'f', '1'] and "
                                                            "['R', 'r', '2'] respectively.", sequence_file)



import unittest
from unittest.mock import patch
from csv import reader
from io import StringIO

from iridauploader.parsers.miniseq.validation import validate_sample_sheet
from iridauploader.parsers.exceptions import SampleSheetError


class TestValidation(unittest.TestCase):
    """
    Tests valid and invalid sample sheets
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_validate_sample_sheet_no_header(self, mock_csv_reader):
        """
        Given a sample sheet with no header, make sure the correct errors are included in the response
        :param mock_csv_reader:
        :return:
        """
        headers = ("Sample_ID,Sample_Name,"
                   "I7_Index_ID,index,I5_Index_ID,index2,Sample_Project")

        field_values = (
            "15-0318-4004,15-0318,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
            "15-0455-4004,15-0455,N701,TAAGGCGA,S503,TATCCTCT,203\n"
            "15-0462-4004,15-0462,N701,TAAGGCGA,S505,GTAAGGAG,203\n"
        )

        reads = (
            "151\n"
            "151\n"
        )

        file_contents_str = (
            "[Reads]\n"
            "{reads}\n"
            "[Data]\n"
            "{headers}\n"
            "{field_values}"
        ).format(headers=headers, reads=reads, field_values=field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        res = validate_sample_sheet(None)

        # This should be an invalid sample sheet
        self.assertFalse(res.is_valid())
        # Only should have 1 error
        self.assertEqual(len(res.error_list), 1)
        # Error type should be SampleSheetError
        self.assertEqual(type(res.error_list[0]), SampleSheetError)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_validate_sample_sheet_no_data(self, mock_csv_reader):
        """
        Given a sample sheet with no data, make sure the correct errors are included in the response
        :param mock_csv_reader:
        :return:
        """
        field_values = (
            "15-0318-4004,15-0318,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
            "15-0455-4004,15-0455,N701,TAAGGCGA,S503,TATCCTCT,203\n"
            "15-0462-4004,15-0462,N701,TAAGGCGA,S505,GTAAGGAG,203\n"
            "Local Run Manager, 4004\n"
            "Experiment Name, Some_Test_Data\n"
            "Date, 2015-05-14\n"
            "Workflow, GenerateFastQWorkflow\n"
            "Description, 12-34\n"
            "Chemistry, Yes\n"
        )

        reads = (
            "151\n"
            "151\n"
        )

        file_contents_str = (
            "[Header]\n"
            "{field_values}\n"
            "[Reads]\n"
            "{reads}"
        ).format(field_values=field_values, reads=reads)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        res = validate_sample_sheet(None)

        # This should be an invalid sample sheet
        self.assertFalse(res.is_valid())
        # Only should have 2 error
        self.assertEqual(len(res.error_list), 2)
        # Error type should be SampleSheetError
        self.assertEqual(type(res.error_list[0]), SampleSheetError)
        self.assertEqual(type(res.error_list[1]), SampleSheetError)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_validate_sample_sheet_missing_data_header(self, mock_csv_reader):
        """
        Given a sample sheet with no data header, make sure the correct errors are included in the response
        :param mock_csv_reader:
        :return:
        """
        h_field_values = (
            "Local Run Manager Analysis Id, 4004\n"
            "Experiment Name, Some_Test_Data\n"
            "Date, 2015-05-14\n"
            "Workflow, GenerateFastQWorkflow\n"
            "Description, 12-34\n"
            "Chemistry, Yes\n"
        )

        reads = (
            "151\n"
            "151\n"
        )

        d_headers = ("Sample_Name,"
                     "I7_Index_ID,index,I5_Index_ID,index2,Sample_Project,")

        d_field_values = (
            "15-0318-4004,15-0318,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
            "15-0455-4004,15-0455,N701,TAAGGCGA,S503,TATCCTCT,203\n"
            "15-0462-4004,15-0462,N701,TAAGGCGA,S505,GTAAGGAG,203\n"
        )

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values, reads=reads, d_headers=d_headers, d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        res = validate_sample_sheet(None)

        # This should be an invalid sample sheet
        self.assertFalse(res.is_valid())
        # Only should have 1 error
        self.assertEqual(len(res.error_list), 1)
        # Error type should be SampleSheetError
        self.assertEqual(type(res.error_list[0]), SampleSheetError)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_validate_sample_sheet_valid(self, mock_csv_reader):
        """
        Given a valid sample sheet, test that everything shows as valid
        :param mock_csv_reader:
        :return:
        """
        h_field_values = (
            "Local Run Manager Analysis Id, 4004\n"
            "Experiment Name, Some_Test_Data\n"
            "Date, 2015-05-14\n"
            "Workflow, GenerateFastQWorkflow\n"
            "Description, 12-34\n"
            "Chemistry, Yes\n"
        )

        reads = (
            "151\n"
            "151\n"
        )

        d_headers = ("Sample_ID,Sample_Name,"
                     "I7_Index_ID,index,I5_Index_ID,index2,Sample_Project,")

        d_field_values = (
            "15-0318-4004,15-0318,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
            "15-0455-4004,15-0455,N701,TAAGGCGA,S503,TATCCTCT,203\n"
            "15-0462-4004,15-0462,N701,TAAGGCGA,S505,GTAAGGAG,203\n"
        )

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values, reads=reads, d_headers=d_headers, d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        res = validate_sample_sheet(None)

        # This should be a valid sample sheet
        self.assertTrue(res.is_valid())



class SampleSheetError(Exception):
    """An exception raised when errors are encountered with a sample sheet.

    Examples include when a sample sheet can't be parsed because it's garbled, or
    if IRIDA rejects the creation of a run because fields are missing or invalid
    from the sample sheet.
    """
    def __init__(self, message, sample_sheet):
        """Initialize a SampleSheetError.

        Args:
            message: a summary message that's causing the error.
            sample_sheet: the full file path of the sample sheet that has the error
        """
        self._message = message
        self._sample_sheet = sample_sheet

    @property
    def message(self):
        return self._message

    @property
    def errors(self):
        return self._sample_sheet

    def __str__(self):
        return self.message



import unittest
from unittest.mock import patch
from os import path


from iridauploader.core import file_size_validator
from iridauploader import model

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestValidateFileSizeMinimum(unittest.TestCase):
    """
    Testing the validate_file_size_minimum function
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @staticmethod
    def _make_seq_run():
        """
        Make a sequencing run pointed at real data for the tests
        :return: SequencingRun object
        """
        files_1 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R2_001.fastq.gz"),
        ])
        files_2 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222_S1_L001_R2_001.fastq.gz"),
        ])
        files_3 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R2_001.fastq.gz"),
        ])
        sample_1 = model.Sample("test_sample", "description", 1)
        sample_1.sequence_file = files_1
        sample_2 = model.Sample("test_sample", "description", 1)
        sample_2.sequence_file = files_2
        sample_3 = model.Sample("test_sample", "description", 1)
        sample_3.sequence_file = files_3
        project = model.Project("test_project", [sample_1, sample_2, sample_3], "description")
        sequencing_run = model.SequencingRun({"layoutType": "PAIRED_END"}, [project], "miseq")
        return sequencing_run

    @patch("iridauploader.core.file_size_validator.config.read_config_option")
    def test_files_too_small(self, mock_read_file_size_config_option):
        # force the function to grab 100 (min file size) when it tries to call config
        mock_read_file_size_config_option.side_effect = [100]

        # make a sequencing run to work with
        sequencing_run = self._make_seq_run()

        # Run code to test
        res = file_size_validator.validate_file_size_minimum(sequencing_run)

        # validate result
        self.assertFalse(res.is_valid(), "File size is not too small")
        print(res)

    @patch("iridauploader.core.file_size_validator.config.read_config_option")
    def test_files_not_too_small(self, mock_read_file_size_config_option):
        # force the function to grab 0 (min file size) when it tries to call config
        mock_read_file_size_config_option.side_effect = [0]

        # make a sequencing run to work with
        sequencing_run = self._make_seq_run()

        # Run code to test
        res = file_size_validator.validate_file_size_minimum(sequencing_run)

        # validate result
        self.assertTrue(res.is_valid(), "File size is too small")
        print(res)



"""
A Sample will store (key: value) pairs using a dictionary.

Keys from a sequencer include: 'sampleName','description'

Keys from Irida will include these AND many others
"""
from cerberus import Validator, TypeDefinition
from copy import deepcopy

from iridauploader.model.sequence_file import SequenceFile


class Sample:

    # Define SequenceFile as a type for validation
    _sample_type = TypeDefinition('sequence_file', (SequenceFile,), ())
    Validator.types_mapping['sequence_file'] = _sample_type

    uploadable_schema = {
        '_sequence_file': {
            'type': 'sequence_file',
            'nullable': False,
            'required': True,
        },
        '_sample_name': {
            'type': 'string',
            'nullable': False,
            'required': True,
            'minlength': 3  # Minimum sample name length is 3
        },
        '_description': {
            'type': 'string',
            'nullable': True,
            'required': False
        },
        '_sample_number': {
            'anyof_type': ['string', 'integer'],
            'nullable': True,
            'required': False
        },
        '_sample_id': {
            'type': 'integer',
            'nullable': True,
            'required': False
        },
        '_skip': {
            'type': 'boolean',
            'nullable': True,
            'required': False
        }
    }

    def __init__(self, sample_name, description='', sample_number=None, samp_dict=None, sample_id=None):
        """
        :param sample_name: string: displayed sample name on IRIDA
        :param description: string:
        :param sample_number: string or int: used during parsing step for some parsers that define their own numbers for samples
        :param samp_dict: dictionary of additional values
        :param sample_id: int: unique identifier defined by irida
        """
        self._sample_name = sample_name
        self._description = description
        self._sample_number = sample_number
        self._sample_id = sample_id
        if samp_dict is None:
            samp_dict = {}
        self._sample_dict = dict(samp_dict)
        self._sequence_file = None
        self._skip = False

    @property
    def sample_name(self):
        return self._sample_name

    @property
    def description(self):
        return self._description

    @property
    def sample_number(self):
        return self._sample_number

    @property
    def sample_id(self):
        return self._sample_id

    @property
    def sequence_file(self):
        return self._sequence_file

    @sequence_file.setter
    def sequence_file(self, sq):
        self._sequence_file = sq

    @property
    def sample_dict(self):
        return self._sample_dict

    @property
    def skip(self):
        return self._skip

    @skip.setter
    def skip(self, skip):
        self._skip = skip

    def get_uploadable_dict(self):  # formatting for sending to irida when creating a project
        uploadable_dict = deepcopy(self._sample_dict)
        uploadable_dict['sampleName'] = self.sample_name
        uploadable_dict['description'] = self.description
        return uploadable_dict

    def __getitem__(self, key):
        if key in self._sample_dict:
            return self._sample_dict[key]
        return None

    def get(self, key):
        return self.__getitem__(key)

    def __str__(self):
        d = self.get_uploadable_dict()
        d["sequenceFile"] = str(self.sequence_file)
        return str(d)

    def get_dict(self):
        return self.__dict__



from appdirs import user_log_dir
import os
import logging.handlers


# Normal base logging directory name
log_directory_name = "irida-uploader"
# When running tests, the Makefile creates an environment variable IRIDA_UPLOADER_TEST to 'True'
# If it exists then we are running a test and should be logging to the test logs directory
if os.environ.get('IRIDA_UPLOADER_TEST'):
    log_directory_name = "irida_uploader_test"
# Use systems default logging path, and append our named directory
log_file_path = os.path.join(user_log_dir(log_directory_name), 'irida-uploader.log')

if not os.path.exists(user_log_dir(log_directory_name)):
    os.makedirs(user_log_dir(log_directory_name))

# Looks something like this:
# 2019-02-07 14:50:02 INFO     Log message goes here...
log_format = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

# setup root logger
root_logger = logging.getLogger()
root_logger.handlers = []
logging.basicConfig(
    level=logging.NOTSET,  # Default to highest (NOTSET) level, so everything is possible to be logged by handlers
    handlers=[logging.NullHandler()]  # Default log to Null, so that we can handle it manually
)

# Log to file
rotating_file_handler = logging.handlers.RotatingFileHandler(
    filename=log_file_path,
    maxBytes=(1024 * 1024 * 1024 * 10),  # 10GB max file size
    backupCount=100,
)
rotating_file_handler.setLevel(logging.DEBUG)
rotating_file_handler.setFormatter(log_format)
root_logger.addHandler(rotating_file_handler)

# Log to the user
console = logging.StreamHandler()
console.setLevel(logging.INFO)
console.setFormatter(log_format)
root_logger.addHandler(console)

# manages the logging directory
# only one directory can have a logger at a time
directory_logger = None


def add_log_to_directory(directory):
    """
    Starts up a logging handler that creates a log file in the directory being uploaded

    :param directory: directory to create a logger in
    :return: None
    """
    global directory_logger

    # If there is already a directory logger in place, throw an exception
    if directory_logger:
        logging.error("A directory logger already exists!")
        raise Exception("ERROR:add_log_to_directory: A directory logger already exists!")

    logging.info("Adding log file to {}".format(directory))
    log_file = os.path.join(directory, 'irida-uploader.log')
    directory_logger = logging.handlers.RotatingFileHandler(
        filename=log_file,
        maxBytes=(1024 * 1024 * 1024 * 10),  # 10GB max file size
        backupCount=100,
    )
    directory_logger.setLevel(logging.INFO)
    directory_logger.setFormatter(log_format)
    root_logger.addHandler(directory_logger)


def remove_directory_logger():
    """
    Deletes the existing directory logger so logging stops

    :return: None
    """
    global directory_logger
    root_logger.removeHandler(directory_logger)
    directory_logger = None
    logging.info("Stopped active logging to run directory")


def get_user_log_dir():
    return user_log_dir(log_directory_name)



import unittest
import os
from unittest.mock import patch


from iridauploader.parsers import common
from iridauploader.parsers.exceptions import SampleSheetError, DirectoryError

path_to_module = os.path.abspath(os.path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestFindDirectoryList(unittest.TestCase):
    """
    Test getting the list of directories
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_find_three(self):
        """
        Given a directory with 3 run directories in it, make sure all 3 directories are included in result
        :return:
        """
        directory = os.path.join(path_to_module, "three_dirs")
        dir_1 = os.path.join(directory, "first")
        dir_2 = os.path.join(directory, "second")
        dir_3 = os.path.join(directory, "third")
        res = common.find_directory_list(directory)

        self.assertIn(dir_1, res)
        self.assertIn(dir_2, res)
        self.assertIn(dir_3, res)

    def test_find_none(self):
        """
        Given a directory with no sequencing run directories in it, make sure an empty list is returned
        :return:
        """
        directory = os.path.join(path_to_module, "no_dirs")
        res = common.find_directory_list(directory)

        self.assertEqual(res, [])

    @patch("iridauploader.parsers.common.cannot_read_directory")
    def test_cannot_read_directory(self, mock_cannot_read_directory):
        """
        ensure exception is raised when an invalid directory is given
        """
        directory = os.path.join(path_to_module, "not a dir")

        mock_cannot_read_directory.side_effect = [True]

        with self.assertRaises(DirectoryError):
            res = common.find_directory_list(directory)


class TestGetCsvReader(unittest.TestCase):
    """
    Test that the csv reader behaves as expected
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_get_csv_reader_valid_sheet(self):
        """
        Given a valid sample sheet, ensure the parsed sheet matches expected output
        :return:
        """
        sheet_file = os.path.join(path_to_module, "test_csv_reader.csv")

        lines = common.get_csv_reader(sheet_file)
        # This is a sample of what the miseq sample sheet looks like, but it also makes a good
        # example for what we want our csv reader to be able to parse.
        correct_lines = [
            ['[Header]'],
            ['IEMFileVersion', '4'],
            ['Investigator Name', 'Some Guy'],
            ['Experiment Name', '1'],
            ['Date', '10/15/2013'],
            ['Workflow', 'GenerateFASTQ'],
            ['Application', 'FASTQ Only'],
            ['Assay', 'Nextera XT'],
            ['Description', 'Superbug'],
            ['Chemistry', 'Amplicon'],
            [],
            ['[Reads]'],
            ['251'],
            ['250'],
            [],
            ['[Settings]'],
            ['ReverseComplement', '0'],
            ['Adapter', 'AAAAGGGGAAAAGGGGAAA'],
            [],
            ['[Data]'],
            ['Sample_ID', 'Sample_Name', 'Sample_Plate', 'Sample_Well', 'I7_Index_ID', 'index', 'I5_Index_ID', 'index2',
             'Sample_Project', 'Description'],
            ['01-1111', '01-1111', '1', '01', 'N01', 'AAAAAAAA', 'S01', 'TTTTTTTT', '6', 'Super bug '],
            ['02-2222', '02-2222', '2', '02', 'N02', 'GGGGGGGG', 'S02', 'CCCCCCCC', '6', 'Scary bug '],
            ['03-3333', '03-3333', '3', '03', 'N03', 'CCCCCCCC', 'S03', 'GGGGGGGG', '6', 'Deadly bug ']
        ]

        for line, c_line in zip(lines, correct_lines):
            self.assertEqual(line, c_line)

    def test_get_csv_reader_no_sheet(self):
        """
        When no sheet is given to parser, throw error
        :return:
        """
        sheet_file = os.path.join(path_to_module, "fake_dir_data")

        with self.assertRaises(SampleSheetError):
            common.get_csv_reader(sheet_file)



import iridauploader.model as model
from iridauploader.parsers.exceptions.sequence_file_error import SequenceFileError


def validate_uniform_file_count(sequencing_run):
    """
    Validate the files in a SequencingRun object are all either in pairs, or single files

    :param sequencing_run: SequencingRun object to validate
    :return: ValidationResult object with list of errors if any
    """
    paired = sequencing_run.is_paired_end()
    expected_file_count = 2 if paired else 1

    validation_result = model.ValidationResult()

    for p in sequencing_run.project_list:
        for s in p.sample_list:
            # do validation of files
            if not _matching_find_count(s, paired):
                error_msg = "File count for sample `{}` does not match the expected file count `{}`. " \
                            "Please verify your data.".format(s.sample_name, expected_file_count)
                validation_result.add_error(SequenceFileError(error_msg))

    return validation_result


def _matching_find_count(sample, paired):
    """
    checks paired end / single end file matching on sample

    :param sample: Sample object
    :param paired: boolean
    :return: boolean
    """
    return sample.sequence_file.is_paired_end() == paired
